"""Module that contains all parsers related to LArCV particle data.

Contains the following parsers:
- :class:`ParticleParser`
- :class:`NeutrinoParser`
- :class:`ParticlePointParser`
- :class:`ParticleCoordinateParser`
- :class:`ParticleGraphParser`
- :class:`ParticlePIDParser`
- :class:`ParticleEnergyParser`
"""

import numpy as np

from spine.data import Meta, Neutrino, Particle
from spine.utils.conditional import larcv
from spine.utils.globals import (
    INVAL_ID,
    PDG_TO_PID,
    PID_MASSES,
    PPN_LPART_COL,
    TRACK_SHP,
    VALUE_COL,
)
from spine.utils.gnn.network import filter_invalid_nodes
from spine.utils.particles import process_particles
from spine.utils.ppn import get_ppn_labels, get_vertex_labels, image_coordinates

from .base import ParserBase
from .data import ParserObjectList, ParserTensor

__all__ = [
    "ParticleParser",
    "NeutrinoParser",
    "ParticlePointParser",
    "ParticleCoordinateParser",
    "VertexPointParser",
    "ParticleGraphParser",
    "SingleParticlePIDParser",
    "SingleParticleEnergyParser",
]


class ParticleParser(ParserBase):
    """Class which loads larcv.Particle objects to local Particle ones.

    .. code-block. yaml

        schema:
          particles:
            parser: particle
            particle_event: particle_pcluster
            cluster_event: cluster3d_pcluster
            asis: False
            pixel_coordinates: True
            post_process: True
    """

    # Name of the parser (as specified in the configuration)
    name = "particle"

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __init__(
        self,
        pixel_coordinates=True,
        post_process=True,
        skip_empty=False,
        asis=False,
        **kwargs,
    ):
        """Initialize the parser.

        Parameters
        ----------
        pixel_coordinates : bool, default True
            If set to `True`, the parser rescales the truth positions
            (start, end, etc.) to voxel coordinates
        post_process : bool, default True
            Processes particles to add/correct missing attributes
        skip_empty : bool, default False
            Do not read the truth information corresponding to empty particles.
            This saves considerable read time when there are a lot of irrelevant
            particle stored in the LArCV file. It puts an empty `Particle`
            object in place of empty particles, to preserve list size and typing.
        asis : bool, default False
            Load the objects as larcv objects, do not build local data class
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.pixel_coordinates = pixel_coordinates
        self.post_process = post_process
        self.skip_empty = skip_empty
        self.asis = asis

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(
        self,
        particle_event,
        sparse_event=None,
        cluster_event=None,
        particle_mpv_event=None,
        neutrino_event=None,
    ):
        """Fetch the list of true particle objects.

        Parameters
        ----------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles
        sparse_event : larcv.EventSparseTensor3D, optional
            Tensor which contains the metadata needed to convert the
            positions in voxel coordinates
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster which contains the metadata needed to convert the
            positions in voxel coordinates
        particle_mpv_event : larcv.EventParticle, optional
            Particle event which contains the list of true MPV particles
        neutrino_event : larcv.EventNeutrino, optional
            Neutrino event which contains the list of true neutrinos

        Returns
        -------
        List[Particle]
            List of true particle objects
        """
        # If asis is true, return larcv objects
        particle_list = list(particle_event.as_vector())
        if self.asis:
            assert (
                not self.pixel_coordinates
            ), "If `asis` is True, `pixel_coordinates` must be False."
            assert (
                not self.post_process
            ), "If `asis` is True, `post_process` must be False."
            assert (
                not self.skip_empty
            ), "If `asis` is True`, `skip_empty` must be False."

            return ParserObjectList(particle_list, larcv.Particle())

        # Convert to a list of particle objects
        particles = []
        for p in particle_list:
            if not self.skip_empty or p.num_voxels() > 0 or p.id() == p.group_id():
                particles.append(Particle.from_larcv(p))
            else:
                particles.append(Particle())

        # If requested, post-process the particle list
        if self.post_process:
            process_particles(
                particles, particle_event, particle_mpv_event, neutrino_event
            )

        # If requested, convert the point positions to pixel coordinates
        if self.pixel_coordinates:
            # Fetch the metadata
            assert (sparse_event is not None) ^ (cluster_event is not None), (
                "Must provide either `sparse_event` or `cluster_event` to "
                "get the metadata and convert positions to voxel units."
            )
            ref_event = sparse_event if sparse_event is not None else cluster_event
            meta = Meta.from_larcv(ref_event.meta())

            # Convert all the relevant attributes
            for p in particles:
                if p.id > -1:
                    p.to_px(meta)

        # Define the shifts to be applied to each index attribute
        num_particles = len(particles)
        if neutrino_event is not None:
            num_neutrinos = len(neutrino_event.as_vector())
        else:
            nu_ids = np.array([part.nu_id for part in particles], dtype=int)
            num_neutrinos = np.max(nu_ids, initial=-1) + 1

        index_shifts = {}
        for attr in Particle().index_attrs:
            index_shifts[attr] = num_particles if attr != "nu_id" else num_neutrinos

        # Return
        return ParserObjectList(particles, Particle(), index_shifts)


class NeutrinoParser(ParserBase):
    """Class which loads larcv.Neutrino objects to local Neutrino ones.

    .. code-block. yaml

        schema:
          neutrinos:
            parser: neutrino
            neutrino_event: neutrino_mpv
            cluster_event: cluster3d_pcluster
            pixel_coordinates: True
            asis: False
    """

    # Name of the parser (as specified in the configuration)
    name = "neutrino"

    # Type of object(s) returned by the parser
    returns = "object_list"

    def __init__(self, pixel_coordinates=True, asis=False, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        pixel_coordinates : bool, default True
            If set to `True`, the parser rescales the truth positions
            (start, end, etc.) to voxel coordinates
        asis : bool, default False
            Load the objects as larcv objects, do not build local data class
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.pixel_coordinates = pixel_coordinates
        self.asis = asis

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, neutrino_event, sparse_event=None, cluster_event=None):
        """Fetch the list of true neutrino objects.

        Parameters
        ----------
        neutrino_event : larcv.EventNeutrino
            Neutrino event which contains the list of true neutrinos
        sparse_event : larcv.EventSparseTensor3D, optional
            Tensor which contains the metadata needed to convert the
            positions in voxel coordinates
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster which contains the metadata needed to convert the
            positions in voxel coordinates

        Returns
        -------
        List[Neutrino]
            List of true neutrino objects
        """
        # If asis is true, return larcv objects
        neutrino_list = list(neutrino_event.as_vector())
        if self.asis:
            assert (
                not self.pixel_coordinates
            ), "If `asis` is True, `pixel_coordinates` must be False."

            return ParserObjectList(neutrino_list, larcv.Neutrino())

        # Convert to a list of neutrino objects
        neutrinos = [Neutrino.from_larcv(n) for n in neutrino_list]

        # If requested, convert the point positions to pixel coordinates
        if self.pixel_coordinates:
            # Fetch the metadata
            assert (sparse_event is not None) ^ (cluster_event is not None), (
                "Must provide either `sparse_event` or `cluster_event` to "
                "get the metadata and convert positions to voxel units."
            )
            ref_event = sparse_event if sparse_event is not None else cluster_event
            meta = Meta.from_larcv(ref_event.meta())

            # Convert all the relevant attributes
            for n in neutrinos:
                n.to_px(meta)

        # Define the shifts to be applied to each index attribute
        inter_ids = [n.interaction_id for n in neutrinos]
        num_neutrinos = len(neutrino_event.as_vector())
        max_inter = np.max(inter_ids, initial=-1) + 1
        index_shifts = {"id": num_neutrinos, "interaction_id": max_inter}

        return ParserObjectList(neutrinos, Neutrino(), index_shifts)


class ParticlePointParser(ParserBase):
    """Class that retrieves the points of interests.

    It provides the coordinates of the end points, types and particle index.

    .. code-block. yaml

        schema:
          points:
            parser: particle_points
            particle_event: particle_pcluster
            sparse_event: sparse3d_pcluster
            include_point_tagging: True
    """

    # Name of the parser (as specified in the configuration)
    name = "particle_points"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(self, include_point_tagging=True, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        include_point_tagging : bool, default True
            If `True`, includes start vs end point tagging
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.include_point_tagging = include_point_tagging

        # Define the output columns containing indexes
        self.index_cols = np.array([PPN_LPART_COL])

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, particle_event, sparse_event=None, cluster_event=None):
        """Fetch the list of label points of interest.

        Parameters
        ----------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles
        sparse_event : larcv.EventSparseTensor3D, optional
            Tensor which contains the metadata needed to convert the
            positions in voxel coordinates
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster which contains the metadata needed to convert the
            positions in voxel coordinates

        Returns
        -------
        np_voxels : np.ndarray
            (N, 3) array of [x, y, z] coordinates
        np_features : np.ndarray
            (N, 2/3) array of [point type, particle index(, end point tagging)]
        meta : Meta
            Metadata of the parsed image
        """
        # Fetch the metadata
        assert (sparse_event is not None) ^ (cluster_event is not None), (
            "Must provide either `sparse_event` or `cluster_event` to "
            "get the metadata and convert positions to voxel units."
        )
        ref_event = sparse_event if sparse_event is not None else cluster_event
        meta = ref_event.meta()

        # Get the point labels
        particle_v = particle_event.as_vector()
        point_labels = get_ppn_labels(
            particle_v,
            meta,
            self.ftype,
            include_point_tagging=self.include_point_tagging,
        )

        return ParserTensor(
            coords=point_labels[:, :3],
            features=point_labels[:, 3:],
            meta=Meta.from_larcv(meta),
            index_cols=self.index_cols,
            index_shifts=np.array([particle_event.size()]),
        )


class ParticleCoordinateParser(ParserBase):
    """Class that retrieves that end points of particles.

    It provides the coordinates of the end points, time and shape.

    .. code-block. yaml

        schema:
          coords:
            parser: particle_coordinates
            particle_event: particle_pcluster
            sparse_event: sparse3d_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "particle_coords"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, particle_event, sparse_event=None, cluster_event=None):
        """Fetch the start/end point and time of each true particle.

        Parameters
        ----------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles
        sparse_event : larcv.EventSparseTensor3D, optional
            Tensor which contains the metadata needed to convert the
            positions in voxel coordinates
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster which contains the metadata needed to convert the
            positions in voxel coordinates

        Returns
        -------
        np_voxels : np.ndarray
            (N, 6) array of [x_s, y_s, z_s, x_e, y_e, z_e] start and end
            point coordinates
        np_features : np.ndarray
            (N, 2) array of [first_step_t, shape_id]
        meta : Meta
            Metadata of the parsed image
        """
        # Fetch the metadata
        assert (sparse_event is not None) ^ (cluster_event is not None), (
            "Must provide either `sparse_event` or `cluster_event` to "
            "get the metadata and convert positions to voxel units."
        )
        ref_event = sparse_event if sparse_event is not None else cluster_event
        meta = ref_event.meta()

        # Scale particle coordinates to image size
        particle_v = particle_event.as_vector()

        # Make features
        coord_labels = np.empty((len(particle_v), 8), dtype=self.ftype)
        for i, p in enumerate(particle_v):
            start_point = last_point = image_coordinates(meta, p.first_step())
            if p.shape() == TRACK_SHP:  # End point only meaningful for tracks
                last_point = image_coordinates(meta, p.last_step())
            extra = [p.t(), p.shape()]
            coord_labels[i] = np.concatenate((start_point, last_point, extra))

        return ParserTensor(
            coords=coord_labels[:, :6],
            features=coord_labels[:, 6:],
            meta=Meta.from_larcv(meta),
        )


class VertexPointParser(ParserBase):
    """Class that retrieves the vertices.

    It provides the coordinates of points where multiple particles originate:
    - If the `neutrino_event` is provided, it simply uses the coordinates of
      the neutrino interaction points.
    - If the `particle_event` is provided instead, it looks for ancestor point
      positions shared by at least two particles.

    .. code-block. yaml

        schema:
          vertices:
            parser: vertex_points
            particle_event: particle_pcluster
            #neutrino_event: neutrino_mpv
            sparse_event: sparse3d_pcluster
            include_point_tagging: True
    """

    # Name of the parser (as specified in the configuration)
    name = "vertex_points"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(self, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Define the output columns containing indexes
        self.index_cols = np.array([VALUE_COL])

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(
        self,
        particle_event=None,
        neutrino_event=None,
        sparse_event=None,
        cluster_event=None,
    ):
        """Fetch the list of label vertex points.

        Parameters
        ----------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles
        neutrino_event : larcv.EventNeutrino
            Neutrino event which contains the list of true neutrinos
        sparse_event : larcv.EventSparseTensor3D, optional
            Tensor which contains the metadata needed to convert the
            positions in voxel coordinates
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster which contains the metadata needed to convert the
            positions in voxel coordinates

        Returns
        -------
        np_voxels : np.ndarray
            (N, 3) array of [x, y, z] coordinates
        np_features : np.ndarray
            (N, 1) array of [vertex ID]
        meta : Meta
            Metadata of the parsed image
        """
        # Check that only one source of vertex is provided
        assert (particle_event is not None) ^ (neutrino_event is not None), (
            "Must provide either `particle_event` or `sparse_event` to "
            "get the vertex points, not both."
        )

        # Fetch the metadata
        assert (sparse_event is not None) ^ (cluster_event is not None), (
            "Must provide either `sparse_event` or `cluster_event` to "
            "get the metadata and convert positions to voxel units."
        )
        ref_event = sparse_event if sparse_event is not None else cluster_event
        meta = ref_event.meta()

        # Get the vertex labels
        particle_v = particle_event.as_vector() if particle_event else None
        neutrino_v = neutrino_event.as_vector() if neutrino_event else None
        point_labels = get_vertex_labels(particle_v, neutrino_v, meta, self.ftype)

        # Get the index shift to apply
        index_shifts = np.max(point_labels[:, 0], keepdims=True, initial=-1) + 1

        return ParserTensor(
            coords=point_labels[:, :3],
            features=point_labels[:, 3:],
            meta=Meta.from_larcv(meta),
            index_cols=self.index_cols,
            index_shifts=index_shifts,
        )


class ParticleGraphParser(ParserBase):
    """Class that uses larcv.EventParticle to construct edges
    between particles (i.e. clusters).

    .. code-block. yaml

        schema:
          graph:
            parser: particle_graph
            particle_event: particle_pcluster
            cluster_event: cluster3d_pcluster
            include_fragment_edges: false
    """

    # Name of the parser (as specified in the configuration)
    name = "particle_graph"

    # Type of object(s) returned by the parser
    returns = "tensor"

    def __init__(self, include_fragment_edges=False, **kwargs):
        """Initialize the parser.

        Parameters
        ----------
        include_fragment_edges : bool, default False
            If `True`, includes edges which join particles in the same group
        **kwargs : dict, optional
            Data product arguments to be passed to the `process` function
        """
        # Initialize the parent class
        super().__init__(**kwargs)

        # Store the revelant attributes
        self.include_fragment_edges = include_fragment_edges

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, particle_event, cluster_event=None):
        """Fetch the parentage connections from the true particle list.

        Configuration
        -------------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles
        cluster_event : larcv.EventClusterVoxel3D, optional
            Cluster used to check if particles have 0 pixel in the image. If
            so, the edges to those clusters are removed and the broken
            parantage is subsequently patched.

        Returns
        -------
        np.ndarray
            (2, E) Array of directed edges for each [parent, child] connection
        int
            Number of particles in the input
        """
        # Check on the cluster input, if provided
        particles_v = particle_event.as_vector()
        num_particles = particles_v.size()
        if cluster_event is not None:
            # Check that the cluster list is of the expected length
            clusters_v = cluster_event.as_vector()
            num_clusters = len(clusters_v)
            assert num_particles == num_clusters or num_particles == num_clusters - 1, (
                f"The number of particles ({num_particles}) must be "
                f"aligned with the number of clusters ({num_clusters}). "
                "There can me one more catch-all cluster at the end."
            )

        # Build a list of edges
        edges, zero_nodes = [], []
        for cluster_id, part in enumerate(particles_v):
            # If the parent ID is invalid (broken parentage), skip
            if part.parent_id() == INVAL_ID:
                continue

            # Only include edges within particle groups if explicitely requested
            if not self.include_fragment_edges and part.group_id() != cluster_id:
                continue

            # Add edge between particle and its direct parent
            if part.parent_id() != part.group_id():
                edges.append([int(part.parent_id()), cluster_id])

            # If the cluster event is provided, keep track of empty nodes
            if cluster_event is not None and clusters_v[cluster_id].size() == 0:
                zero_nodes.append(cluster_id)

        # Convert the list of edges to a numpy array
        if not edges:
            edges = np.empty((0, 2), dtype=np.int64)
        else:
            edges = np.vstack(edges).astype(np.int64)

        # Remove zero-pixel nodes, if possible
        if len(zero_nodes) > 0 and len(edges) > 0:
            edges = filter_invalid_nodes(edges, zero_nodes)

        return ParserTensor(features=edges.T, global_shift=num_particles)


class SingleParticlePIDParser(ParserBase):
    """Get the first true particle's species.

    .. code-block. yaml

        schema:
          image_pid:
            parser: single_particle_pid
            particle_event: particle_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "single_particle_pid"

    # Type of object(s) returned by the parser
    returns = "scalar"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, particle_event):
        """Fetch the species of the first particle.

        Configuration
        -------------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles

        Returns
        -------
        int
            Species of the first particle
        """
        pid = -1
        for p in particle_event.as_vector():
            if p.track_id() == 1:
                if int(p.pdg_code()) in PDG_TO_PID.keys():
                    pid = PDG_TO_PID[int(p.pdg_code())]

                break

        return pid


class SingleParticleEnergyParser(ParserBase):
    """Get the first true particle's kinetic energy.

    .. code-block. yaml

        schema:
          image_energy:
            parser: single_particle_energy
            particle_event: particle_pcluster
    """

    # Name of the parser (as specified in the configuration)
    name = "single_particle_energy"

    # Type of object(s) returned by the parser
    returns = "scalar"

    # Overlay strategy for the objects returned by the parser
    overlay = "cat"

    def __call__(self, trees):
        """Parse one entry.

        Parameters
        ----------
        trees : dict
            Dictionary which maps each data product name to a LArCV object
        """
        return self.process(**self.get_input_data(trees))

    def process(self, particle_event):
        """Fetch the kinetic energy of the first particle.

        Configuration
        -------------
        particle_event : larcv.EventParticle
            Particle event which contains the list of true particles

        Returns
        -------
        float
            Kinetic energy of the first particle
        """
        ke = -1.0
        for p in particle_event.as_vector():
            if p.track_id() == 1:
                if int(p.pdg_code()) in PDG_TO_PID.keys():
                    einit = p.energy_init()
                    pid = PDG_TO_PID[int(p.pdg_code())]
                    mass = PID_MASSES[pid]
                    ke = einit - mass

                break

        return ke



import networkx as nx
import numba as nb
import numpy as np
import torch

from .shapley import mc_shapley


class MCTSNode:
    """
    Node class for Monte Carlo Tree Search in SubgraphX.

    Attributes:
        - graph: the (sub)graph representing the MCTS node
        - actions: the set of all actions that can be taken at this MCTS node.
        - W:
    """

    def __init__(self, G: nx.Graph, eps: float = 0.01):
        self.graph = G
        self.num_particles = len(G.nodes)
        self.actions = np.zeros(self.num_particles)
        self.W = np.zeros(self.num_particles)
        self.C = np.zeros(self.num_particles)
        self.P = np.zeros(self.num_particles)
        self.eps = eps

    def Q(self):
        return self.W / self.C

    def U(self, a):
        factor = np.sqrt(np.sum(self.C)) / (1 + np.sum(self.C))
        return self.eps * self.R(a) * factor

    def R(self, a):
        return mc_shapley()


class MCTS:

    pass


class SubgraphX:

    def __init__(self, model, G, num_mcts, leaf_threshold):

        pass



"""MST graph constructor for GNNs."""

import numba as nb
import numpy as np
from scipy.sparse.csgraph import minimum_spanning_tree

from spine.math.linalg import submatrix

from .base import GraphBase

__all__ = ["MSTGraph"]


class MSTGraph(GraphBase):
    """Generates graphs based on the minimum-spanning tree (MST) of the input
    node locations.

    Makes an edge for each branch in the minimum-spanning tree.

    See :class:`GraphBase` for attributes/methods shared
    across all graph constructors.
    """

    # Name of the graph constructor (as specified in the configuration)
    name = "mst"

    def generate(self, clusts, dist_mat, **kwargs):
        """Generates an incidence matrix that connects nodes that share an
        edge in their corresponding Euclidean MST graph.

        Parameters
        ----------
        clusts : IndexBatch
            (C) Cluster indexes
        dist_mat : Union[np.ndarray, torch.Tensor]
            (C, C) Matrix of pair-wise distances between clusters in the batch
        **kwargs : dict, optional
            Unused graph generation arguments

        Returns
        -------
        np.ndarray
            (2, E) Tensor of edges
        """
        return self._generate(clusts.batch_ids, dist_mat, self.directed)

    @staticmethod
    @nb.njit(cache=True)
    def _generate(
        batch_ids: nb.int64[:], dist_mat: nb.float64[:, :], directed: bool = False
    ) -> nb.int64[:, :]:
        # For each batch, find the list of edges, append it
        edge_list = []
        num_edges = 0
        ret = np.empty((0, 2), dtype=np.int64)
        for b in np.unique(batch_ids):
            clust_ids = np.where(batch_ids == b)[0]
            if len(clust_ids) > 1:
                submat = np.triu(submatrix(dist_mat, clust_ids, clust_ids))
                # Suboptimal. Ideally want to reimplement in Numba, tall order.
                with nb.objmode(mst_mat="float32[:,:]"):
                    mst_mat = minimum_spanning_tree(submat)
                    mst_mat = mst_mat.toarray().astype(np.float32)
                edges = np.where(mst_mat > 0.0)
                edges = np.vstack((clust_ids[edges[0]], clust_ids[edges[1]])).T
                ret = np.vstack((ret, edges))

        return ret.T



"""Module to write log files to CSV."""

import os

__all__ = ["CSVWriter"]


class CSVWriter:
    """Writes data to a CSV file.

    Builds a CSV file to store the output of the analysis tools. It can only be
    used to store relatively basic quantities (scalars, strings, etc.).

    Typical configuration should look like:

    .. code-block:: yaml

        io:
          ...
          writer:
            name: csv
            file_name: output.csv
    """

    name = "csv"

    def __init__(
        self,
        file_name="output.csv",
        overwrite=False,
        append=False,
        accept_missing=False,
    ):
        """Initialize the basics of the output file.

        Parameters
        ----------
        file_name : str, default 'output.csv'
            Name of the output CSV file
        overwrite : bool, default False
            If True, overwrite the output file if it already exists
        append : bool, default False
            If True, add more rows to an existing CSV file
        accept_missing : bool, default True
            Tolerate missing keys
        """
        # Check that output file does not already exist, if requestes
        if not overwrite and os.path.isfile(file_name):
            raise FileExistsError(f"File with name {file_name} already exists.")

        # Store persistent attributes
        self.file_name = file_name
        self.append_file = append
        self.accept_missing = accept_missing
        self.result_keys = None
        if self.append_file:
            if not os.path.isfile(file_name):
                raise FileNotFoundError(
                    f"File not found at path: {file_name}. When using "
                    "`append=True` in CSVWriter, the file must exist at "
                    "the prescribed path before data is written to it."
                )

            with open(self.file_name, "r", encoding="utf-8") as out_file:
                self.result_keys = out_file.readline().split(",")

    def create(self, result_blob):
        """Initialize the header of the CSV file, record the keys to be stored.

        Parameters
        ----------
        result_blob : dict
            Dictionary containing the output of the reconstruction chain
        """
        # Save the list of keys to store
        self.result_keys = list(result_blob.keys())

        # Create a header and write it to file
        with open(self.file_name, "w", encoding="utf-8") as out_file:
            header_str = ",".join(self.result_keys)
            out_file.write(header_str + "\n")

    def append(self, result_blob):
        """Append the CSV file with the output.

        Parameters
        ----------
        result_blob : dict
            Dictionary containing the output of the reconstruction chain
        """
        # Fetch the values to store
        if self.result_keys is None:
            # If this function has never been called, initialiaze the CSV file
            self.create(result_blob)

        else:
            # If it has, check that the list of keys is identical
            if list(result_blob.keys()) != self.result_keys:
                # If it is not identical, check the discrepancies
                missing = self.array_diff(self.result_keys, result_blob.keys())
                excess = self.array_diff(result_blob.keys(), self.result_keys)
                if len(excess):
                    raise AssertionError(
                        "There are keys in this entry which were not "
                        "present when the CSV file was initialized. "
                        f"New keys: {list(excess)}"
                    )

                if not self.accept_missing:
                    raise AssertionError(
                        "There are keys missing in this entry which were "
                        "present when the CSV file was initialized. "
                        f"Missing keys: {list(missing)}"
                    )

                new_result_blob = {k: -1 for k in self.result_keys}
                for k, v in result_blob.items():
                    new_result_blob[k] = v
                result_blob = new_result_blob

        # Append file
        with open(self.file_name, "a", encoding="utf-8") as out_file:
            result_str = ",".join([str(result_blob[k]) for k in self.result_keys])
            out_file.write(result_str + "\n")

    @staticmethod
    def array_diff(array_x, array_y):
        """Compare the content of two arrays.

        This functions returns the elemnts of the first array that
        do not appear in the second array.

        Parameters
        ----------
        array_x : List[str]
            First array of strings
        array_y : List[str]
            Second array of strings

        Returns
        -------
        Set[str]
            Set of keys that appear in `array_x` but not in `array_y`.
        """
        return set(array_x).difference(set(array_y))



"""Data structure construction and analysis object building.

This module handles the construction of high-level analysis data classes from
ML reconstruction outputs and provides I/O capabilities for analysis workflows.

**Note**: The configuration block is still named `build:` for backward compatibility,
while the module is named `construct` to avoid git ignore conflicts.

**Core Construction Management:**
- `BuildManager`: Central factory for analysis object construction
  - Configuration-driven building pipeline
  - Truth-reconstruction matching
  - Quality validation and filtering
  - Batch processing capabilities

**Analysis Data Classes:**

**Fragment-Level Objects:**
- `RecoFragment`: Reconstructed cluster/fragment with ML predictions
- `TruthFragment`: Monte Carlo truth fragment with simulation info
  - Spatial coordinates and energy depositions
  - Particle association and containment
  - Reconstruction quality metrics
  - Truth matching and overlap calculations

**Particle-Level Objects:**
- `RecoParticle`: Reconstructed particle with kinematics and properties
- `TruthParticle`: Monte Carlo truth particle with full genealogy
  - 4-momentum and trajectory information
  - Particle identification and confidence
  - Parent-child relationships
  - Detector interaction points

**Interaction-Level Objects:**
- `RecoInteraction`: Complete reconstructed neutrino interaction
- `TruthInteraction`: Monte Carlo truth interaction with primaries
  - Primary vertex and topology
  - All associated particles and fragments
  - Event-level quantities and classifications
  - Cross-detector information (CRT, PMT)

**Core Functionality:**

**Object Construction:**
- **From ML Chain**: Build objects from raw model predictions
  - Segmentation and clustering results
  - Particle identification scores
  - Point proposal network outputs
  - Graph neural network predictions

**Data Loading:**
- **From HDF5**: Load pre-built objects from analysis files
  - Efficient hierarchical data access
  - Lazy loading for large datasets
  - Metadata preservation and validation
  - Cross-reference integrity checking

**Quality Control:**
- Object validation and consistency checks
- Truth matching quality assessment
- Reconstruction completeness metrics
- Error flagging and diagnostic info

**Data Hierarchy:**
```
ML Outputs → Build Process → Analysis Objects → Physics Analysis
    ↓            ↓              ↓                ↓
Predictions   Construction   Fragments         Performance
Clusters      Matching       Particles         Studies
Scores        Validation     Interactions      Results
```

**Object Relationships:**
- **Hierarchical Structure**: Interactions → Particles → Fragments
- **Truth Matching**: Reco objects linked to MC truth counterparts
- **Cross-References**: Bi-directional navigation between related objects
- **Inheritance**: Shared base classes with automatic docstring merging

**I/O Capabilities:**
- **HDF5 Export**: Structured storage with compression and metadata
- **Batch Processing**: Multi-file dataset construction
- **Memory Management**: Efficient loading for large-scale analysis
- **Format Validation**: Schema checking and version compatibility

**Key Features:**
- Type-safe object construction
- Automatic truth-reco association
- Comprehensive object validation
- Flexible configuration system
- Memory-efficient data structures
- Rich metadata preservation
- Cross-platform compatibility

**Example Usage:**
```python
from spine.construct import BuildManager

# Build objects from ML chain output
builder = BuildManager(config)
objects = builder.build_from_ml_chain(ml_results)

# Load objects from analysis file
objects = builder.load_from_file('analysis.h5')

# Access hierarchy
interaction = objects.interactions[0]
particles = interaction.particles  # List of RecoParticle
fragments = particles[0].fragments  # List of RecoFragment
```

This module provides the foundation for all downstream physics analysis
by creating well-structured, validated analysis objects with full provenance.
"""

from .manager import BuildManager



"""Manages the operation of post-processors."""

from collections import OrderedDict, defaultdict
from copy import deepcopy

import numpy as np

from spine.utils.stopwatch import StopwatchManager

from .factories import post_processor_factory


class PostManager:
    """Manager in charge of handling post-processing scripts.

    It loads all the post-processor objects once and feeds them data.
    """

    def __init__(self, cfg, post_list=None, parent_path=None):
        """Initialize the post-processing manager.

        Parameters
        ----------
        cfg : dict
            Post-processor configurations
        post_list : List[str], optional
            List of post-processors which have already been run
        parent_path : str, optional
            Path to the analysis tools configuration file
        """
        # Loop over the post-processor modules and get their priorities
        cfg = deepcopy(cfg)
        keys = np.array(list(cfg.keys()))
        priorities = -np.ones(len(keys), dtype=np.int32)
        for i, key in enumerate(keys):
            if "priority" in cfg[key]:
                priorities[i] = cfg[key].pop("priority")

        # Add the modules to a processor list in decreasing order of priority
        self.watch = StopwatchManager()
        self.modules = OrderedDict()
        keys = keys[np.argsort(-priorities)]
        for key in keys:
            # Profile the module
            self.watch.initialize(key)

            # Append
            self.modules[key] = post_processor_factory(
                key, cfg[key], parent_path=parent_path
            )

            # Check dependencies
            if post_list is not None:
                ups_post = tuple(self.modules)
                for post in self.modules[key]._upstream:
                    assert post in (post_list + ups_post), (
                        f"Post-processor `{key}` is missing an essential "
                        f"upstream post-processor: `{post}`."
                    )

    def __call__(self, data):
        """Pass one batch of data through the post-processors.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over the post-processor modules
        single_entry = np.isscalar(data["index"])
        for key, module in self.modules.items():
            # Run the post-processor on each entry
            self.watch.start(key)
            if single_entry:
                num_entries = 1
                result = module(data)

            else:
                num_entries = len(data["index"])
                result = defaultdict(list)
                for entry in range(num_entries):
                    result_e = module(data, entry)
                    if result_e is not None:
                        for k, v in result_e.items():
                            result[k].append(v)

            self.watch.stop(key)

            # Update the input dictionary
            if result is not None:
                for key, val in result.items():
                    if not single_entry:
                        assert len(val) == num_entries, (
                            f"The number of {key} ({len(val)}) returned by the {key} "
                            "post-processor does not match the number of entries "
                            f"({num_entries}) in the batch."
                        )
                    data[key] = val



"""YAML loader with SPINE-specific features.

This module provides the ConfigLoader class which extends yaml.SafeLoader
with support for:
- !include tag for inline file includes
- !path tag for path resolution
- !download tag for downloading files from URLs
"""

import os
from typing import Any, List, Optional, TextIO, Union, cast

import yaml

from .download import download_from_url
from .errors import ConfigIncludeError

__all__ = ["ConfigLoader", "resolve_config_path"]


def resolve_config_path(
    filename: str, current_dir: str, search_paths: Optional[List[str]] = None
) -> str:
    """Resolve a configuration file path with SPINE_CONFIG_PATH support.

    Resolution order:
    1. If absolute path, return as-is
    2. Try relative to current_dir
    3. Try relative to current_dir with .yaml/.yml extension
    4. Search through SPINE_CONFIG_PATH directories
    5. Search through SPINE_CONFIG_PATH directories with .yaml/.yml extension

    Parameters
    ----------
    filename : str
        Config filename or path to resolve
    current_dir : str
        Directory of the config file doing the including
    search_paths : Optional[List[str]]
        List of search paths (defaults to SPINE_CONFIG_PATH env var)

    Returns
    -------
    str
        Resolved absolute path

    Raises
    ------
    ConfigIncludeError
        If file cannot be found in any location
    """
    # If absolute path, check if it exists
    if os.path.isabs(filename):
        if os.path.exists(filename):
            return filename
        raise ConfigIncludeError(f"Absolute path not found: {filename}")

    # Try relative to current directory
    relative_path = os.path.join(current_dir, filename)
    if os.path.exists(relative_path):
        return os.path.abspath(relative_path)

    # Try with .yaml/.yml extension relative to current directory
    for ext in [".yaml", ".yml"]:
        if not filename.endswith(ext):
            relative_path_with_ext = relative_path + ext
            if os.path.exists(relative_path_with_ext):
                return os.path.abspath(relative_path_with_ext)

    # Get search paths from environment variable if not provided
    if search_paths is None:
        env_paths = os.environ.get("SPINE_CONFIG_PATH", "")
        search_paths = [p.strip() for p in env_paths.split(":") if p.strip()]

    # Search through SPINE_CONFIG_PATH
    for search_dir in search_paths:
        search_path = os.path.join(search_dir, filename)
        if os.path.exists(search_path):
            return os.path.abspath(search_path)

        # Try with extensions
        for ext in [".yaml", ".yml"]:
            if not filename.endswith(ext):
                search_path_with_ext = search_path + ext
                if os.path.exists(search_path_with_ext):
                    return os.path.abspath(search_path_with_ext)

    # File not found anywhere
    search_locations = [f"Relative to: {current_dir}"]
    if search_paths:
        search_locations.append(f"SPINE_CONFIG_PATH: {':'.join(search_paths)}")

    raise ConfigIncludeError(
        f"Config file '{filename}' not found.\n"
        f"Searched in:\n  - " + "\n  - ".join(search_locations)
    )


class ConfigLoader(yaml.SafeLoader):
    """YAML loader with !include tag support.

    This loader extends yaml.SafeLoader to support inline file includes
    using the !include tag.
    """

    def __init__(
        self, stream: Union[TextIO, str], root_dir: Optional[str] = None
    ) -> None:
        """Initialize the loader.

        Parameters
        ----------
        stream : Union[TextIO, str]
            File stream (from `open()`) or string content
        root_dir : Optional[str]
            Root directory for resolving relative paths.
            If None and stream is a file, uses the file's directory.
        """
        if isinstance(stream, str):
            # String input
            self._root = root_dir if root_dir is not None else os.getcwd()
        else:
            # File stream
            if root_dir is not None:
                self._root = root_dir
            else:
                self._root = os.path.split(stream.name)[0]
        super().__init__(stream)

    def include(self, node: yaml.Node) -> Any:
        """Load and include a YAML file inline.

        Parameters
        ----------
        node : yaml.Node
            YAML node containing the filename

        Returns
        -------
        Any
            Loaded configuration content
        """
        filename = self.construct_scalar(cast(yaml.ScalarNode, node))
        resolved_path = resolve_config_path(filename, self._root)

        with open(resolved_path, "r", encoding="utf-8") as f:
            # Create a loader class with the resolved directory as root_dir
            resolved_dir = os.path.dirname(resolved_path)

            class NestedConfigLoader(ConfigLoader):
                def __init__(self, stream):
                    super().__init__(stream, resolved_dir)

            return yaml.load(f, Loader=NestedConfigLoader)

    def resolve_path(self, node: yaml.Node) -> str:
        """Resolve a file path relative to the current config file.

        This is useful for paths that need to be resolved at load time but
        not included as configuration (e.g., model weights, data files, etc.).

        Unlike !include which loads the content, this just resolves the path
        and verifies the file exists.

        Parameters
        ----------
        node : yaml.Node
            YAML node containing the filename

        Returns
        -------
        str
            Resolved absolute path

        Raises
        ------
        ConfigIncludeError
            If file not found

        Examples
        --------
        post:
          flash_match:
            cfg: !path flashmatch/config.yaml  # Resolved relative to this config
        model:
          weights: !path weights/model.ckpt  # Must exist at load time
        """
        filename = self.construct_scalar(cast(yaml.ScalarNode, node))
        resolved_path = resolve_config_path(filename, self._root)
        return resolved_path

    def download(self, node: yaml.Node) -> str:
        """Download a file from URL and return the cached path.

        This is useful for downloading model weights or other large files
        that should not be stored in git but need to be accessible.

        Files are cached in a weights/ directory (or SPINE_CACHE_DIR if set).
        If a file already exists in cache, it is not re-downloaded.

        Parameters
        ----------
        node : yaml.Node
            YAML node containing the URL (string or dict with url/hash)

        Returns
        -------
        str
            Absolute path to cached file

        Raises
        ------
        HTTPError
            If download fails
        ValueError
            If hash validation fails

        Examples
        --------
        model:
          # Simple URL
          weights: !download https://example.com/model.ckpt

          # With hash validation
          weights: !download
            url: https://example.com/model.ckpt
            hash: abc123def456...  # SHA256 hash
        """
        if isinstance(node, yaml.ScalarNode):
            # Simple case: just a URL string
            url = self.construct_scalar(node)
            return download_from_url(url)
        elif isinstance(node, yaml.MappingNode):
            # Complex case: dict with url and optional hash
            data = self.construct_mapping(node)
            url = data.get("url")
            expected_hash = data.get("hash")

            if not url:
                raise ConfigIncludeError(
                    "!download requires either a URL string or a dict with 'url' key"
                )

            return download_from_url(url, expected_hash=expected_hash)
        else:
            raise ConfigIncludeError(
                f"!download expects a string URL or dict, got {type(node)}"
            )


# Register the !include, !path, and !download constructors
ConfigLoader.add_constructor("!include", ConfigLoader.include)
ConfigLoader.add_constructor("!path", ConfigLoader.resolve_path)
ConfigLoader.add_constructor("!download", ConfigLoader.download)



"""Edge kernel functions that produce edge weights from node features."""

import torch
import torch.nn as nn

from spine.model.layer.common.mlp import MLP

__all__ = ["DefaultKernel", "MixedKernel", "BilinearKernel", "MLPKernel"]


class DefaultKernel(nn.Module):
    """Kernel producing edge score based on feature L2 similarity.

    This Kernel assumes that the upstream embedder produces a set of spatial
    and embedding coordinates and computes the L2 similarity between the two
    node feature vectors. It scales the L2 distance by the covariance and
    penalizes for cluster size dissimilarity.
    """

    name = "default"

    def __init__(self, num_features, eps=1e-3):
        """Initializes the kernel.

        Parameters
        ----------
        num_features : int
            Number of dimensions in feature embedding space
        eps : float
            Features regularization factor
        """
        # Initialize the parent class
        super().__init__()

        # Store the parameters
        self.num_features = num_features
        self.eps = eps

    def compute_edge_weight(
        self, sp_emb1, sp_emb2, ft_emb1, ft_emb2, cov1, cov2, occ1, occ2
    ):
        """Converts the output of the embedder into an edge score.

        Parameters
        ----------
        sp_emb1 : torch.Tensor
            (E, 3) Spatial embeddings of the source nodes
        sp_emb2 : torch.Tensor
            (E, 3) Spatial embeddings of the target nodes
        ft_emb1 : torch.Tensor
            (E, N_f) Feature embeddings of the source nodes
        ft_emb2 : torch.Tensor
            (E, N_f) Feature embeddings of the target nodes
        cov1 : torch.Tensor
            (E, 2) Spatial extent of the cluster the source nodes belongs to
        cov2 : torch.Tensor
            (E, 2) Spatial extent of the cluster the target nodes belongs to
        occ1 : torch.Tensor
            (E, 1) Multiplicity of the cluster the source nodes belongs to
        occ2 : torch.Tensor
            (E, 1) Multiplicity of the cluster the target nodes belongs to

        Returns
        -------
        torch.Tensor
            Scalar value of the edge weight
        """
        # Measure spatial distance between nodes, weighted by cluster covariance
        sp_cov_i = cov1[:, 0]
        sp_cov_j = cov2[:, 0]
        sp_i = ((sp_emb1 - sp_emb2) ** 2).sum(dim=1) / (sp_cov_i**2 + self.eps)
        sp_j = ((sp_emb1 - sp_emb2) ** 2).sum(dim=1) / (sp_cov_j**2 + self.eps)

        # Measure feature distance between nodes, weighted by cluster covariance
        ft_cov_i = cov1[:, 1]
        ft_cov_j = cov2[:, 1]
        ft_i = ((ft_emb1 - ft_emb2) ** 2).sum(dim=1) / (ft_cov_i**2 + self.eps)
        ft_j = ((ft_emb1 - ft_emb2) ** 2).sum(dim=1) / (ft_cov_j**2 + self.eps)

        # Convert the L2 distances to a probability measure (Gaussian kernel)
        p_ij = torch.exp(-sp_i - ft_i)
        p_ji = torch.exp(-sp_j - ft_j)

        pvec = torch.clamp(p_ij + p_ji - p_ij * p_ji, min=0, max=1)

        # Scale down the probability if the cluster sizes are highly different
        r1, r2 = occ1.flatten(), occ2.flatten()
        r = torch.max(
            (r2 + self.eps) / (r1 + self.eps), (r1 + self.eps) / (r2 + self.eps)
        )
        pvec /= r

        return pvec

    def forward(self, x1, x2):
        """Computes the kernel edge score of ll node pairs in the graph.

        This kernel expectes a set of (3 + N_f + 2 + 1) features per node:
        - 3 spatial embedding features
        - N_f feature embedding features
        - 2 covariance features
        - 1 occupancy feature

        Parameters
        ----------
        x1 : torch.Tensor
            (E, 3 + N_f + 2 + 1) Features of the source nodes
        x2 : torch.Tensor
            (E, 3 + N_f + 2 + 1) Features of the target nodes
        """
        # Decompose the two feature sets into their consituents
        assert (
            x1.shape[1] == x2.shape[1] == 6 + self.num_features
        ), "The combined feature vector is not of the expected shape."

        nf = self.num_features
        splits = [3, nf + 3, nf + 5]
        sp_emb1, ft_emb1, cov1, occ1 = torch.tensor_split(x1, splits, dim=1)
        sp_emb2, ft_emb2, cov2, occ2 = torch.tensor_split(x1, splits, dim=1)

        # Compute the edge weight, make sure it's between 0 and 1
        weight = self.compute_edge_weight(
            sp_emb1, sb_emb2, ft_emb1, ft_emb2, cov1, cov2, occ1, occ2
        )
        weight = torch.clamp(w, min=0 + 1e-6, max=1 - 1e-6)

        # Convert probability to a logit, return
        result = torch.logit(weight)

        return result


class MixedKernel(DefaultKernel):
    """Kernel producing edge score based on feature L2 similarity and cosine
    similarity between node features.

    This Kernel assumes that the upstream embedder produces a set of spatial
    and embedding coordinates and computes the L2 similarity between the two
    node feature vectors. It scales the L2 distance by the covariance and
    penalizes for cluster size dissimilarity.

    In addition, it computes cosine similarity between the feature vectors.
    """

    name = "mixed"

    def __init__(self, num_features, eps=1e-3):
        """Initializes the kernel.

        Parameters
        ----------
        num_features : int
            Number of dimensions in feature embedding space
        eps : float
            Features regularization factor
        """
        # Initialize the parent class
        super().__init__(num_features, eps)

        # Initialize the cosine similarity layer
        self.cos = nn.CosineSimilarity(dim=1)

    def compute_edge_weight_coord(
        self, coord1, coord2, tan1, tan2, coord_cov1, coord_cov2, tan_cov1, tan_cov2
    ):
        """Converts the the spatial coordinates information into an edge score.

        Parameters
        ----------
        coord1 : torch.Tensor
            (E, 3) Coordinates of the source nodes
        coord2 : torch.Tensor
            (E, 3) Coordinates of the target nodes
        tan1 : torch.Tensor
            (E, N_f) Tangents of the source nodes
        tan2 : torch.Tensor
            (E, N_f) Tangents of the target nodes
        coord_cov1 : torch.Tensor
            (E, 3) Spatial extent of the cluster the source nodes belongs to
        coord_cov2 : torch.Tensor
            (E, 3) Spatial extent of the cluster the target nodes belongs to
        tan_cov1 : torch.Tensor
            (E, 1) Multiplicity of the cluster the source nodes belongs to
        tan_cov2 : torch.Tensor
            (E, 1) Multiplicity of the cluster the target nodes belongs to

        Returns
        -------
        torch.Tensor
            Scalar value of the edge weight
        """
        # Compute the average position of the two nodes
        coord_cov = (coord_cov1 + coord_cov2) / 2.0

        # Compute the vector which joins the two nodes and its length
        chord = coord1 - coord2
        chord_dist = torch.pow(chord, 2)
        dist = torch.sum(chord_dist * coord_cov, dim=1)

        # Weight is proportial to the distance
        coords_weight = torch.exp(dist)

        # Affinity
        a1 = torch.abs(self.cos(chord, tan1))
        a2 = torch.abs(self.cos(chord, tan2))
        norm_factor = torch.sum(chord_dist, dim=1)
        tan_cov = (tan_cov1 + tan_cov2) / 2.0
        a = a1 * a2 * tan_cov / (norm_factor + 1e-5)
        affinity_weight = torch.exp(-a)

        pvec = coords_weight * affinity_weight

        return pvec

    def forward(self, x1, x2):
        """Computes the kernel edge score of all node pairs in the graph.

        This kernel expectes a set of (3 + 3 + 3 + N_f + 2 + 3 + 1 + 1)
        features per node:
        - 3 coordinates
        - 3 tangent components
        - 3 spatial embedding features
        - 16 feature embedding features
        - 2 covariance features
        - 3 coordinate covariance features
        - 1 tangent coviarance feature
        - 1 occupancy feature

        Parameters
        ----------
        x1 : torch.Tensor
            (E, 3 + 3 + 3 + N_f + 2 + 3 + 1 + 1) Features of the source nodes
        x2 : torch.Tensor
            (E, 3 + 3 + 3 + N_f + 2 + 3 + 1 + 1) Features of the targer nodes
        """
        # Decompose the two feature sets into their consituents
        assert (
            x1.shape[1] == x2.shape[1] == 16 + self.num_features
        ), "The combined feature vector is not of the expected shape."

        nf = self.num_features
        splits = [3, 6, 9, nf + 9, nf + 11, nf + 14, nf + 15]
        coord1, tan1, sp_emb1, ft_emb1, cov1, coord_cov1, tan_cov1, occ1 = (
            torch.tensor_split(x1, splits, dim=1)
        )
        coord2, tan2, sp_emb2, ft_emb2, cov2, coord_cov2, tan_cov2, occ2 = (
            torch.tensor_split(x1, splits, dim=1)
        )

        # Compute the L2 edge weight
        weight1 = self.compute_edge_weight(
            sp_emb1, sb_emb2, ft_emb1, ft_emb2, cov1, cov2, occ1, occ2
        )

        # Compute the cosine edge weight
        weight2 = self.compute_edge_weight_coord(
            coord1, coord2, tan1, tan2, coord_cov1, coord_cov2, tan_cov1, tan_cov2
        )

        # Combine the scroes and make sure they are between 0 and 1
        weight = torch.clamp(w1 * w2, min=0 + 1e-6, max=1 - 1e-6)

        # Convert probability to a logit, return
        result = torch.logit(weight)

        return result


class BilinearKernel(nn.Module):
    """Kernel producing edges scores based on a learnable bilinear layer."""

    name = "bilinear"

    def __init__(self, num_features, bias=False):
        """Initializes the kernel.

        Parameters
        ----------
        num_features : int
            Number of dimensions in feature embedding space
        bias : bool, default False
            If `True`, allows for an overall bias in the bilinear layer
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the bilinear layer
        self.bilin = nn.Bilinear(num_features, num_features, 1, bias=bias)

        # Store parameter
        self.num_features = num_features

    def forward(self, x1, x2):
        """Computes the kernel edge score of all node pairs in the graph.

        Parameters
        ----------
        x1 : torch.Tensor
            (E, N_f) Features of the source nodes
        x2 : torch.Tensor
            (E, N_f) Features of the targer nodes
        """
        # Check on input size, pass through the bilinear layer
        assert (
            x1.shape[1] == x2.shape[1] == self.num_features
        ), "The feature vector is not of the expected shape."

        return self.bilin(x1, x2)


class MLPKernel(nn.Module):
    """Kernel producing edges scores based on an MLP and a linear layer."""

    name = "mlp"

    def __init__(self, num_features, bias=False, mlp=None):
        """Initializes the kernel.

        Parameters
        ----------
        num_features : int
            Number of dimensions in feature embedding space
        bias : bool, default False
            If `True`, allows for an overall bias in the bilinear layer
        mlp : dict, optional
            MLP architecture configuration, see :class:`MLP`
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the underlying MLP
        if mlp is None:
            mlp = {
                "depth": 2,
                "width": 32,
                "activation": "elu",
                "normalization": "batch_norm",
            }

        self.mlp = MLP(in_channels=num_features, **mlp)

        # Initialize the final linear layer
        self.lin = nn.Linear(2 * self.mlp.feature_size, 1, bias=bias)

    def forward(self, x1, x2):
        """Computes the kernel edge score of all node pairs in the graph.

        Parameters
        ----------
        x1 : torch.Tensor
            (E, N_f) Features of the source nodes
        x2 : torch.Tensor
            (E, N_f) Features of the targer nodes
        """
        # Pass the node features through the MLP
        f1 = self.mlp(x1)
        f2 = self.mlp(x2)

        # Concatenate them and pass them through the linear layer, return
        result = self.lin(torch.cat([f1, f2], dim=1))

        return result



"""Module with a data class objects which represent output interactions."""

from dataclasses import dataclass, field
from typing import List
from warnings import warn

import numpy as np

from spine.data.neutrino import Neutrino
from spine.utils.docstring import inherit_docstring
from spine.utils.globals import PID_LABELS, PID_TAGS, SHOWR_SHP

from .base import RecoBase, TruthBase

__all__ = ["RecoInteraction", "TruthInteraction"]


@dataclass(eq=False)
class InteractionBase:
    """Base interaction-specific information.

    Attributes
    ----------
    particles : List[object]
        List of particles that make up the interaction
    primary_particles: List[object]
        List of primary particles associated with the interaction
    particle_ids : np.ndarray
        List of Particle IDs that make up this interaction
    primary_particle_ids : np.ndarray
        List of primary Particle IDs associated with this interaction
    num_particles : int
        Number of particles that make up this interaction
    num_primary_particles : int
        Number of primary particles associated with this interaction
    particle_counts : np.ndarray
        (P) Number of particles of each species in this interaction
    primary_particle_counts : np.ndarray
        (P) Number of primary particles of each species in this interaction
    vertex : np.ndarray
        (3) Coordinates of the interaction vertex
    is_fiducial : bool
        Whether this interaction vertex is inside the fiducial volume
    is_flash_matched : bool
        True if the interaction was matched to an optical flash
    flash_ids : np.ndarray
        (F) Indices of the optical flashes the interaction is matched to
    flash_volume_ids : np.ndarray
        (F) Indices of the optical volumes the flashes where recorded in
    flash_times : np.ndarray
        (F) Times at which the flashes occurred in microseconds
    flash_scores : np.ndarray
        (F) Flash matching quality scores reported for each match
    flash_total_pe : float
        Total number of photoelectrons associated with the flash
    flash_hypo_pe : float
        Total number of photoelectrons expected to be produced by the interaction
    is_crt_matched : bool
        True if any particle in the interaction was matched to a CRT hit
    crt_ids : np.ndarray
        (C) Indices of the CRT hits the interaction is matched to
    crt_times : np.ndarray
        (C) Times at which the CRT hits occurred in microseconds
    crt_scores : np.ndarray
        (C) Quality metric associated with the CRT matches
    topology : str
        String representing the interaction topology
    """

    particles: List[object] = None
    primary_particles: List[object] = None
    particle_ids: np.ndarray = None
    primary_particle_ids: np.ndarray = None
    num_particles: int = None
    num_primary_particles: int = None
    particle_counts: np.ndarray = None
    primary_particle_counts: np.ndarray = None
    vertex: np.ndarray = None
    is_fiducial: bool = False
    is_flash_matched: bool = False
    flash_ids: np.ndarray = None
    flash_volume_ids: np.ndarray = None
    flash_times: np.ndarray = None
    flash_scores: np.ndarray = None
    flash_total_pe: float = -1.0
    flash_hypo_pe: float = -1.0
    is_crt_matched: bool = False
    crt_ids: np.ndarray = None
    crt_times: np.ndarray = None
    topology: str = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("vertex", 3),
        ("particle_counts", len(PID_LABELS) - 1),
        ("primary_particle_counts", len(PID_LABELS) - 1),
    )

    # Variable-length attributes as (key, dtype) pairs
    _var_length_attrs = (
        ("particles", object),
        ("primary_particles", object),
        ("particle_ids", np.int32),
        ("primary_particle_ids", np.int32),
        ("flash_ids", np.int32),
        ("flash_volume_ids", np.int32),
        ("flash_times", np.float32),
        ("flash_scores", np.float32),
        ("crt_ids", np.int32),
        ("crt_times", np.float32),
        ("crt_scores", np.float32),
    )

    # Attributes specifying coordinates
    _pos_attrs = ("vertex",)

    # Boolean attributes
    _bool_attrs = ("is_fiducial", "is_flash_matched", "is_crt_matched")

    # Attributes that must never be stored to file
    _skip_attrs = ("particles", "primary_particles")

    def __str__(self):
        """Human-readable string representation of the interaction object.

        Results
        -------
        str
            Basic information about the interaction properties
        """
        match = self.match_ids[0] if len(self.match_ids) > 0 else -1
        info = (
            f"Interaction(ID: {self.id:<3} "
            f"| Size: {self.size:<5} | Topology: {self.topology:<10} "
            f"| Match: {match:<3})"
        )
        if len(self.particles):
            info += "\n" + len(info) * "-"
            for particle in self.particles:
                info += "\n" + str(particle)

        return info

    def reset_flash_match(self, typed=True):
        """Reset all the flash matching attributes.

        Parameters
        ----------
        typed : bool, default True
            If `True`, the underlying arrays are reset to typed empty arrays
        """
        self.is_flash_matched = False
        self.flash_total_pe = -1.0
        self.flash_type_pe = -1.0
        if typed:
            self.flash_ids = np.empty(0, dtype=np.int32)
            self.flash_volume_ids = np.empty(0, dtype=np.int32)
            self.flash_times = np.empty(0, dtype=np.float32)
            self.flash_scores = np.empty(0, dtype=np.float32)
        else:
            self.flash_ids = []
            self.flash_volume_ids = []
            self.flash_times = []
            self.flash_scores = []

    @property
    def primary_particles(self):
        """List of primary particles associated with this interaction.

        Returns
        -------
        List[obect]
            List of primary Particle objects associated with this interaction
        """
        return [part for part in self.particles if part.is_primary]

    @primary_particles.setter
    def primary_particles(self, primary_particles):
        pass

    @property
    def primary_particle_ids(self):
        """List of primary Particle IDs associated with this interaction.

        Returns
        -------
        np.darray
            List of primary Particle IDs associated with this interaction
        """
        return np.array([part.id for part in self.primary_particles])

    @primary_particle_ids.setter
    def primary_particle_ids(self, primary_particle_ids):
        pass

    @property
    def num_particles(self):
        """Number of particles that make up this interaction.

        Returns
        -------
        int
            Number of particles that make up the interaction instance
        """
        return len(self.particle_ids)

    @num_particles.setter
    def num_particles(self, num_particles):
        pass

    @property
    def num_primary_particles(self):
        """Number of primary particles associated with this interaction.

        Returns
        -------
        int
            Number of particles associated with the interaction instance
        """
        return len(self.primary_particle_ids)

    @num_primary_particles.setter
    def num_primary_particles(self, num_primary_particles):
        pass

    @property
    def particle_counts(self):
        """Number of particles of each PID species in this interaction.

        Returns
        -------
        np.ndarray
            (P) Number of particles of each PID
        """
        counts = np.zeros(len(PID_LABELS) - 1, dtype=int)
        for part in self.particles:
            if part.pid > -1 and part.is_valid:
                counts[part.pid] += 1

        return counts

    @particle_counts.setter
    def particle_counts(self, particle_counts):
        pass

    @property
    def primary_particle_counts(self):
        """Number of primary particles of each PID species in this interaction.

        Returns
        -------
        np.ndarray
            (P) Number of primary particles of each PID
        """
        counts = np.zeros(len(PID_LABELS) - 1, dtype=int)
        for part in self.primary_particles:
            if part.pid > -1 and part.is_valid:
                counts[part.pid] += 1

        return counts

    @primary_particle_counts.setter
    def primary_particle_counts(self, primary_particle_counts):
        pass

    @property
    def is_crt_matched(self):
        """Checks if any particle in the interactionw as matched to a CRT hit.

        Returns
        -------
        bool
            `True` if any of the particle was matched to a CRT hit
        """
        return np.any([part.is_crt_matched for part in self.particles])

    @is_crt_matched.setter
    def is_crt_matched(self, is_crt_matched):
        pass

    @property
    def crt_ids(self):
        """Returns the list of CRT hit IDs matched to this interaction.

        Returns
        -------
        np.ndarray
            (C) List of CRT hit IDs matched to this interaction
        """
        if self.is_crt_matched:
            return np.concatenate([part.crt_ids for part in self.particles])
        else:
            return np.empty(0, dtype=np.int32)

    @crt_ids.setter
    def crt_ids(self, crt_ids):
        pass

    @property
    def crt_times(self):
        """Returns the list of CRT hit times matched to this interaction.

        Returns
        -------
        np.ndarray
            (C) List of CRT hit times matched to this interaction
        """
        if self.is_crt_matched:
            return np.concatenate([part.crt_times for part in self.particles])
        else:
            return np.empty(0, dtype=np.int32)

    @crt_times.setter
    def crt_times(self, crt_times):
        pass

    @property
    def crt_scores(self):
        """Returns the list of quality metrics of CRT hits matched to this interaction.

        Returns
        -------
        np.ndarray
            (C) List of quality metrics of CRT hits matched to this interaction
        """
        if self.is_crt_matched:
            return np.concatenate([part.crt_scores for part in self.particles])
        else:
            return np.empty(0, dtype=np.int32)

    @crt_scores.setter
    def crt_scores(self, crt_scores):
        pass

    @property
    def topology(self):
        """String representing the interaction topology.

        Returns
        -------
        str
            String listing the number of primary particles in this interaction
        """
        topology = ""
        for i, count in enumerate(self.primary_particle_counts):
            if count > 0:
                topology += f"{count}{PID_TAGS[i]}"

        return topology

    @topology.setter
    def topology(self, topology):
        pass

    @classmethod
    def from_particles(cls, particles):
        """Builds an Interaction instance from its constituent Particle objects.

        Parameters
        ----------
        particles : List[ParticleBase]
            List of Particle objects that make up the Interaction

        Returns
        -------
        InteractionBase
            Interaction built from the particle list
        """
        # Construct interaction object
        interaction = cls()

        # Fill unique attributes which must be shared between particles
        unique_attrs = ["is_truth", "units"]
        for attr in unique_attrs:
            assert (
                len(np.unique([getattr(p, attr) for p in particles])) < 2
            ), f"{attr} must be unique in the list of particles."

        # Attach particle list
        interaction.particles = particles
        interaction.particle_ids = np.array([p.id for p in particles])

        # Build long-form attributes
        for attr in cls._cat_attrs:
            val_list = [getattr(p, attr) for p in particles]
            setattr(interaction, attr, np.concatenate(val_list))

        return interaction


@dataclass(eq=False)
@inherit_docstring(RecoBase, InteractionBase)
class RecoInteraction(InteractionBase, RecoBase):
    """Reconstructed interaction information."""

    # Attributes that must never be stored to file
    _skip_attrs = (*RecoBase._skip_attrs, *InteractionBase._skip_attrs)

    # Variable-length attributes
    _var_length_attrs = (
        *RecoBase._var_length_attrs,
        *InteractionBase._var_length_attrs,
    )

    # Boolean attributes
    _bool_attrs = (*RecoBase._bool_attrs, *InteractionBase._bool_attrs)

    def __str__(self):
        """Human-readable string representation of the interaction object.

        Results
        -------
        str
            Basic information about the interaction properties
        """
        return "Reco" + super().__str__()

    @property
    def leading_shower(self):
        """Leading primary shower of this interaction.

        Returns
        -------
        RecoParticle
            Primary shower with the highest kinetic energy
        """
        showers = [part for part in self.primary_particles if part.shape == SHOWR_SHP]
        if len(showers) == 0:
            return None

        return max(showers, key=lambda x: x.ke)


@dataclass(eq=False)
@inherit_docstring(TruthBase, InteractionBase)
class TruthInteraction(Neutrino, InteractionBase, TruthBase):
    """Truth interaction information.

    This inherits all of the attributes of :class:`Interaction`, which contains
    the G4 truth information for the interaction.

    Attributes
    ----------
    nu_id : int
        Index of the neutrino matched to this interaction
    reco_vertex : np.ndarray
        (3) Coordinates of the reconstructed interaction vertex
    """

    nu_id: int = -1
    reco_vertex: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("reco_vertex", 3),
        *Neutrino._fixed_length_attrs,
        *InteractionBase._fixed_length_attrs,
    )

    # Variable-length attributes
    _var_length_attrs = (
        *TruthBase._var_length_attrs,
        *InteractionBase._var_length_attrs,
    )

    # Attributes specifying coordinates
    _pos_attrs = ("reco_vertex", *InteractionBase._pos_attrs, *Neutrino._pos_attrs)

    # Boolean attributes
    _bool_attrs = (*TruthBase._bool_attrs, *InteractionBase._bool_attrs)

    # Attributes that must never be stored to file
    _skip_attrs = (*TruthBase._skip_attrs, *InteractionBase._skip_attrs)

    def __str__(self):
        """Human-readable string representation of the interaction object.

        Results
        -------
        str
            Basic information about the interaction properties
        """
        return "Truth" + super().__str__()

    def attach_neutrino(self, neutrino):
        """Attach neutrino generator information to this interaction.

        Parameters
        ----------
        neutrino : Neutrino
            Neutrino to fetch the attributes from
        """
        # Transfer all the neutrino attributes
        for attr, val in neutrino.as_dict().items():
            if attr != "id":
                setattr(self, attr, val)
            else:
                if neutrino.id != self.nu_id:
                    warn(
                        "The neutrino ID as stored in the larcv.Neutrino "
                        "object does not match its index."
                    )

        # Set the interaction vertex position
        self.vertex = neutrino.position



"""Module to draw cylinders."""

import time

import numpy as np
import plotly.graph_objs as go

__all__ = ["cylinder_traces"]


def cylinder_trace(
    centroid,
    axis,
    height,
    diameter,
    num_samples=10,
    color=None,
    intensity=None,
    hovertext=None,
    showscale=False,
    **kwargs,
):
    """Draw a cylinder centered at a given position.

    Parameters
    ----------
    centroid : np.ndarray
        (3) Centroid of the cylinder
    axis : np.ndarray
        (3) Axis direction of the cylinder
    height : float
        Height of the cylinder
    diameter : float
        Diameter of the cylinder
    num_samples : int, default 10
        Number of points sampled along theta and h in the cylindrical coordinate
        system of the cylinder. A larger number increases the resolution.
    color : Union[str, float], optional
        Color of cylinder
    intensity : Union[int, float], optional
        Color intensity of the cylinder along the colorscale axis
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with the cylinder
    showscale : bool, default False
        If True, show the colorscale of the :class:`plotly.graph_objs.Mesh3d`
    **kwargs : dict, optional
        Additional parameters to pass to the underlying
        :class:`plotly.graph_objs.Mesh3d` object
    """
    # Compute the points on a unit cylinder
    phi = np.linspace(0, 2 * np.pi, num=num_samples)
    h = np.linspace(-0.5, 0.5, num=num_samples)
    phi, h = np.meshgrid(phi, h)
    x = 0.5 * np.cos(phi)
    y = 0.5 * np.sin(phi)
    z = h
    unit_points = np.vstack((x.flatten(), y.flatten(), z.flatten())).T

    # Compute the rotation matrix which aligns the z axis to the cylinder axis
    axis = axis / np.linalg.norm(axis)
    z_axis = np.array([0.0, 0.0, 1.0])
    rotmat = np.eye(3)
    if (axis != z_axis).any():
        v = np.cross(z_axis, axis)
        c = np.dot(z_axis, axis)
        s = np.linalg.norm(v)
        kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        rotmat = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s**2))

    # Compute the scaling vectors for radius and height
    scale = np.diag([diameter, diameter, height])

    # Compute the cylinder points
    cyl_points = centroid + np.dot(unit_points.dot(scale), rotmat)

    # Convert the color provided to a set of intensities, if needed
    if color is not None and not isinstance(color, str):
        assert intensity is None, "Must not provide both `color` and `intensity`."
        intensity = np.full(len(cyl_points), color)
        color = None

    # Update hovertemplate style
    hovertemplate = "x: %{x}<br>y: %{y}<br>z: %{z}"
    if hovertext is not None:
        if not np.isscalar(hovertext):
            hovertemplate += "<br>%{text}"
        else:
            hovertemplate += f"<br>{hovertext}"
            hovertext = None

    # Append Mesh3d object
    return go.Mesh3d(
        x=cyl_points[:, 0],
        y=cyl_points[:, 1],
        z=cyl_points[:, 2],
        color=color,
        intensity=intensity,
        alphahull=0,
        showscale=showscale,
        hovertemplate=hovertemplate,
        **kwargs,
    )


def cylinder_traces(
    centroids,
    axis,
    height,
    diameter,
    color=None,
    hovertext=None,
    cmin=None,
    cmax=None,
    shared_legend=True,
    legendgroup=None,
    showlegend=True,
    name=None,
    **kwargs,
):
    """Function which produces a list of plotly traces of cylinders given a
    list of centroids and one covariance matrix in x, y and z.

    Parameters
    ----------
    centroids : np.ndarray
        (N, 3) Positions of each of the cylinder centroids
    axis : np.ndarray
        (3,) or (N, 3) Axis direction of the cylinders
    height : Union[float, np.ndarray]
        Height of the cylinders
    diameter : Union[float, np.ndarray]
        Diameter of the cylinders
    color : Union[str, np.ndarray], optional
        Color of cylinders or list of color of cylinders
    hovertext : Union[int, str, np.ndarray], optional
        Text associated with every cylinder or each cylinder
    cmin : float, optional
        Minimum value along the color scale
    cmax : float, optional
        Maximum value along the color scale
    shared_legend : bool, default True
        If True, the plotly legend of all ellipsoids is shared as one
    legendgroup : str, optional
        Legend group to be shared between all cylinders
    showlegend : bool, default `True`
        Whether to show legends on not
    name : str, optional
        Name of the trace(s)
    **kwargs : dict, optional
        List of additional arguments to pass to the underlying list of
        :class:`plotly.graph_objs.Mesh3D`

    Returns
    -------
    Union[List[plotly.graph_objs.Mesh3D]]
        Ellipsoid traces
    """
    # Check the parameters
    assert (
        color is None or np.isscalar(color) or len(color) == len(centroids)
    ), "Specify one color for all cylinders, or one color per cylinder."
    assert (
        hovertext is None or np.isscalar(hovertext) or len(hovertext) == len(centroids)
    ), "Specify one hovertext for all cylinders, or one hovertext per cylinder."
    assert axis.shape == (3,) or axis.shape == (
        len(centroids),
        3,
    ), "Specify one axis for all cylinders, or one axis per cylinder."
    assert np.isscalar(height) or len(height) == len(
        centroids
    ), "Specify one height for all cylinders, or one height per cylinder."
    assert np.isscalar(diameter) or len(diameter) == len(
        centroids
    ), "Specify one diameter for all cylinders, or one diameter per cylinder."

    # If one color is provided per cylinder, give an associated hovertext
    if hovertext is None and isinstance(color, (list, tuple, np.ndarray)):
        hovertext = [f"Value: {v:0.3f}" for v in color]

    # If cmin/cmax are not provided, must build them so that all cylinders
    # share the same colorscale range (not guaranteed otherwise)
    if color is not None and isinstance(color, (list, tuple, np.ndarray)):
        if len(color) > 0:
            if cmin is None:
                cmin = np.min(color)
            if cmax is None:
                cmax = np.max(color)

    # If the legend is to be shared, make sure there is a common legend group
    if shared_legend and legendgroup is None:
        legendgroup = "group_" + str(time.time())

    # Loop over the list of cylinder centroids
    traces = []
    col, hov = color, hovertext
    for i, centroid in enumerate(centroids):
        # Fetch the right color/hovertext combination
        if color is not None and isinstance(color, (list, tuple, np.ndarray)):
            col = color[i]
        if hovertext is not None and isinstance(hovertext, (list, tuple, np.ndarray)):
            hov = hovertext[i]

        # If the legend is shared, only draw the legend of the first trace
        if shared_legend:
            showlegend = showlegend and i == 0
            name_i = name
        else:
            name_i = f"{name} {i}"

        # If any of the axis, height or diameter are arrays, fetch the right one
        axis_i = axis
        if len(axis.shape) == 2:
            axis_i = axis[i]
        height_i = height
        if isinstance(height, (list, tuple, np.ndarray)):
            height_i = height[i]
        diameter_i = diameter
        if isinstance(diameter, (list, tuple, np.ndarray)):
            diameter_i = diameter[i]

        # Append list of traces
        traces.append(
            cylinder_trace(
                centroid=centroid,
                axis=axis_i,
                height=height_i,
                diameter=diameter_i,
                contour=None,
                color=col,
                hovertext=hov,
                cmin=cmin,
                cmax=cmax,
                legendgroup=legendgroup,
                showlegend=showlegend,
                name=name_i,
                **kwargs,
            )
        )

    return traces



"""Module which holds all detector component geometries.

This includes:
- :class:`TPCDetector` for a set of organized TPCs
- :class:`OptDetector` for a set of organized light collection detectors
- :class:`CRTDetector` for a set of organized CRT planes
"""

from .base import Box, Plane
from .crt import CRTDetector
from .optical import OptDetector
from .tpc import TPCDetector



"""Shower physics utilities."""

import math
from typing import Tuple, Union

import numba as nb
import numpy as np
from scipy.special import gammaincinv  # pylint: disable=E0611

from .globals import LAR_E_CRIT, LAR_X0, LAR_Z


@nb.njit(cache=True)
def shower_long_profile(
    depth: Union[float, np.ndarray], energy: float, use_gp: bool = False
) -> Union[float, np.ndarray]:
    """Compute the longitudinal shower profile using the Gamma distribution.

    Parameters
    ----------
    depth : Union[float, np.ndarray]
        Depths at which to compute the profile in radiation lengths.
    energy : float
        Energy of the shower in MeV.
    use_gp : bool, default False
        Whether to use the Grindhammer and Peters (2000) parametrization.

    Returns
    -------
    Union[float, np.ndarray]
        The longitudinal shower profile at the specified depths.
    """
    # Parametrize the longitudinal profile in units of radiation length
    t = depth / LAR_X0

    # Compute the shape (a) and scale (b) parameters of the Gamma distribution
    if use_gp:
        a, b = shower_long_params_gp(energy)
    else:
        a, b = shower_long_params_lar(energy)

    # Compute the Gamma distribution profile in the depth(s) requested
    coeff = energy * b / math.gamma(a)
    return coeff * (b * t) ** (a - 1) * np.exp(-b * t)


def shower_long_quantile(energy: float, quantile: float, use_gp: bool = False) -> float:
    """Compute the depth corresponding to a given quantile of the longitudinal shower profile.

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.
    quantile : float
        Quantile to compute (between 0 and 1).
    use_gp : bool, default False
        Whether to use the Grindhammer and Peters (2000) parametrization.

    Returns
    -------
    float
        Depth corresponding to the specified quantile in cm.
    """
    # Compute the shape (a) and scale (b) parameters of the Gamma distribution
    if use_gp:
        a, b = shower_long_params_gp(energy)
    else:
        a, b = shower_long_params_lar(energy)

    # Use the inverse of the regularized incomplete gamma function to find t
    t = gammaincinv(a, quantile) / b  # in units of radiation length

    return t * LAR_X0  # convert to cm


@nb.njit(cache=True)
def shower_long_mode_gp(energy: float) -> float:
    """Compute the depth of the shower maximum, according to Grindhammer and Peters (2000).

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    float
        Depth of the shower maximum in cm.

    Notes
    -----
    Based on the parametrization from Grindhammer and Peters (2000).
    Source: https://arxiv.org/pdf/hep-ex/0001020
    """
    y = energy / LAR_E_CRIT
    t_max = np.log(y) - 0.858  # in units of radiation length

    return t_max * LAR_X0  # convert to cm


@nb.njit(cache=True)
def shower_long_maximum_lar(energy: float) -> float:
    """Compute the depth of the shower maximum.

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    float
        Depth of the shower maximum in cm.

    Notes
    -----
    Based on a fit done to an edep-sim simulation of single electrons in LAr.
    """
    y = energy / LAR_E_CRIT
    t_max = 0.689 * np.log(1 + 0.305 * y**1.68)

    return t_max * LAR_X0  # convert to cm


@nb.njit(cache=True)
def shower_long_params_gp(energy: float) -> Tuple[float, float]:
    """Compute the longitudinal shower profile parameters a and b according
    to Grindhammer and Peters (2000).

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    Tuple[float, float]
        Tuple containing shape parameter a and scale parameter b.

    Notes
    -----
    Based on the parametrization from Grindhammer and Peters (2000).
    Source: https://arxiv.org/pdf/hep-ex/0001020
    """
    y = energy / LAR_E_CRIT
    t_max = np.log(y) - 0.858
    a = 0.21 + (0.492 + 2.38 / LAR_Z) * np.log(y)
    b = (a - 1) / t_max

    return a, b


@nb.njit(cache=True)
def shower_long_params_lar(energy: float) -> Tuple[float, float]:
    """Compute the longitudinal shower profile parameters a and b according
    to a general log-based parametrization of the form:

    t_max = B_t * log(1 + B_t*(E / E_crit)**gamma_t) for t
    a = 1 + B_a * log(1 + B_a(E / E_crit)**gamma_a) for a

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    Tuple[float, float]
        Tuple containing shape parameter a and scale parameter b.

    Notes
    -----
    Based on a fit done to an edep-sim simulation of single electrons in LAr.
    """
    y = energy / LAR_E_CRIT
    t_max = 0.689 * np.log(1 + 0.305 * y**1.68)
    a = 1.4 + 0.083 * np.log(1 + 0.21 * y**7.83)
    b = (a - 1) / t_max

    return a, b


@nb.njit(cache=True)
def shower_angle_profile(
    angle: Union[float, np.ndarray], energy: float
) -> Union[float, np.ndarray]:
    """Compute the angular shower profile using a Stacy parametrization.

    Parameters
    ----------
    angle : Union[float, np.ndarray]
        Angles at which to compute the profile in radians.
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    Union[float, np.ndarray]
        The angular shower profile at the specified angles.
    """
    # Compute the Stacy distribution parameters a, b and c
    a, b, c = shower_angle_params(energy)

    # Compute the Stacy distribution profile at the angle(s) requested
    coeff = c / (b**a * math.gamma(a / c))
    return coeff * angle ** (a - 1) * np.exp(-((angle / b) ** c))


def shower_angle_mode(energy: float) -> float:
    """Compute the mode of the angular shower profile.

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    float
        Mode of the angular shower profile in radians.
    """
    # Compute the Stacy distribution parameters a, b and c
    a, b, c = shower_angle_params(energy)

    # Compute the mode of the distribution
    mode = b * ((a - 1) / c) ** (1 / c)

    return mode  # in radians


def shower_angle_quantile(energy: float, quantile: float) -> float:
    """Compute the angle corresponding to a given quantile of the angular shower profile.

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.
    quantile : float
        Quantile to compute (between 0 and 1).

    Returns
    -------
    float
        Angle corresponding to the specified quantile in radians.
    """
    # Compute the Stacy distribution parameters a, b and c
    a, b, c = shower_angle_params(energy)

    # Use the inverse of the regularized incomplete gamma function to find angle
    angle = b * (gammaincinv(a / c, quantile)) ** (1 / c)

    return angle  # in radians


@nb.njit(cache=True)
def shower_angle_params(energy: float) -> Tuple[float, float, float]:
    """Compute the shower angle parameters a, b and c according to a Stacy
    parametrization of the angular distribution.

    Source: https://arxiv.org/pdf/2303.10226

    Parameters
    ----------
    energy : float
        Energy of the shower in MeV.

    Returns
    -------
    float
        Stacy distribution parameter a.
    float
        Stacy distribution parameter b.
    float
        Stacy distribution parameter c.

    Notes
    -----
    Based on a simple parametrization.
    """
    # Simple parametrization of shower angle distribution parameters
    # t params [ 0.00282092  0.19161785  2.13132419 -0.7097491 ]
    # a params [1.48734863 0.09577318 1.53568148 2.62746278]
    # c params [ 1.21861046 -0.01702748  8.02936571  8.91356063]
    y = energy / LAR_E_CRIT
    t_max = 0.192 * np.log(1 + 2.131 * y**-0.71)
    a = 1.49 + 0.0958 * np.log(1 + 1.53 * y**2.63)
    c = 1.22 - 0.017 * np.log(1 + 8.05 * y**8.94)
    b = t_max * ((a - 1) / c) ** (-1 / c)

    return a, b, c



from typing import Union

import MinkowskiEngine as ME
import numpy as np
import torch
import torch.nn as nn

from .act_norm import act_factory, norm_factory


class ConvolutionBlock(ME.MinkowskiNetwork):
    """Convolution block which operates a sequence of
    two (convolution + nomalization + activation) steps.
    """

    def __init__(
        self,
        in_features,
        out_features,
        stride=1,
        dilation=1,
        dimension=3,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the convolution block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        stride : int, default 1
            Convolution kernel stride
        dilation : int, default 1
            Convolution kernel dilation
        dimension : int, default 3
            Dimension of the input image
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert dimension > 0

        self.conv1 = ME.MinkowskiConvolution(
            in_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.norm1 = norm_factory(normalization, out_features)
        self.act_fn1 = act_factory(activation)

        self.conv2 = ME.MinkowskiConvolution(
            out_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.norm2 = norm_factory(normalization, out_features)
        self.act_fn2 = act_factory(activation)

    def forward(self, x):
        """Pass a tensor through the convolution block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        out = self.conv1(x)
        out = self.norm1(out)
        out = self.act_fn1(out)

        out = self.conv2(out)
        out = self.norm2(out)
        out = self.act_fn2(out)

        return out


class DropoutBlock(ME.MinkowskiNetwork):
    """Convolution block which operates a sequence of
    two (convolution + dropout + nomalization + activation) steps.
    """

    def __init__(
        self,
        in_features,
        out_features,
        stride=1,
        dilation=1,
        dimension=3,
        p=0.5,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the dropout block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        stride : int, default 1
            Convolution kernel stride
        dilation : int, default 1
            Convolution kernel dilation
        p : float, default 0.5
            Dropout probability
        dimension : int, default 3
            Dimension of the input image
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert dimension > 0

        self.conv1 = ME.MinkowskiConvolution(
            in_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.dropout1 = ME.MinkowskiDropout(p=p)
        self.norm1 = norm_factory(normalization, out_features)
        self.act_fn1 = act_factory(activation)

        self.conv2 = ME.MinkowskiConvolution(
            out_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.dropout2 = ME.MinkowskiDropout(p=p)
        self.norm2 = norm_factory(normalization, out_features)
        self.act_fn2 = act_factory(activation)

    def forward(self, x):
        """Pass a tensor through the dropout block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        out = self.conv1(x)
        out = self.dropout1(out)
        out = self.norm1(out)
        out = self.act_fn1(out)

        out = self.conv2(out)
        out = self.dropout2(out)
        out = self.norm2(out)
        out = self.act_fn2(out)

        return out


class ResNetBlock(ME.MinkowskiNetwork):
    """ResNet Block."""

    def __init__(
        self,
        in_features,
        out_features,
        stride=1,
        dilation=1,
        dimension=3,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the ResNet block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        stride : int, default 1
            Convolution kernel stride
        dilation : int, default 1
            Convolution kernel dilation
        dimension : int, default 3
            Dimension of the input image
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert dimension > 0

        if in_features != out_features:
            self.residual = ME.MinkowskiLinear(in_features, out_features, bias=bias)
        else:
            self.residual = nn.Identity()

        self.conv1 = ME.MinkowskiConvolution(
            in_features,
            out_features,
            kernel_size=3,
            stride=stride,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.norm1 = norm_factory(normalization, in_features)
        self.act_fn1 = act_factory(activation)

        self.conv2 = ME.MinkowskiConvolution(
            out_features,
            out_features,
            kernel_size=3,
            stride=stride,
            dilation=dilation,
            dimension=dimension,
            bias=bias,
        )
        self.norm2 = norm_factory(normalization, out_features)
        self.act_fn2 = act_factory(activation)

    def forward(self, x):
        """Pass a tensor through the ResNet block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        residual = self.residual(x)

        out = self.conv1(self.act_fn1(self.norm1(x)))
        out = self.conv2(self.act_fn2(self.norm2(out)))

        out += residual

        return out


class AtrousIIBlock(ME.MinkowskiNetwork):
    """ResNet-type block with atrous convolutions.

    Developed for the ACNN paper: "ACNN: a Full Resolution DCNN for Medical
    Image Segmentation"

    Original paper: https://arxiv.org/pdf/1901.09203.pdf
    """

    def __init__(
        self,
        in_features,
        out_features,
        dimension=3,
        activation="relu",
        normalization="batch_norm",
    ):
        """Initialize the AtrousII block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        dimension : int, default 3
            Dimension of the input image
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert dimension > 0

        self.D = dimension

        if in_features != out_features:
            self.residual = ME.MinkowskiLinear(in_features, out_features)
        else:
            self.residual = nn.Identity()

        self.conv1 = ME.MinkowskiConvolution(
            in_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=1,
            dimension=self.D,
        )
        self.norm1 = norm_factory(normalization, out_features)
        self.act_fn1 = act_factory(activation)

        self.conv2 = ME.MinkowskiConvolution(
            out_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=3,
            dimension=self.D,
        )
        self.norm2 = norm_factory(normalization, out_features)
        self.act_fn2 = act_factory(activation)

    def forward(self, x):
        """Pass a tensor through the AtrousII block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        residual = self.residual(x)

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.act_fn1(out)

        out = self.conv2(out)
        out = self.norm2(out)
        out += residual
        out = self.act_fn2(out)

        return out


class ResNeXtBlock(ME.MinkowskiNetwork):
    """ResNeXt-type block with atrous convolutions.

    Notes
    -----
    For vanilla resnext blocks, set `dilation=1` and others to default.
    """

    def __init__(
        self,
        in_features,
        out_features,
        dimension=3,
        cardinality=4,
        depth=1,
        dilations=None,
        kernel_sizes=3,
        strides=1,
        activation="relu",
        normalization="batch_norm",
    ):
        """Initialize the ResNeXt block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        dimension : int, default 3
            Dimension of the input image
        cardinality : int, default 4
            Number of different paths, see ResNeXt paper
        depth : int, default 1
            Number of (convolutions + normalization + activation) layers
        dilations : int, optional
            Dilation rates for each convolution layer inside the cardinal paths
        kernel_sizes : int, default 3
            Kernel sizes for each convolution layer inside the cardinal paths
        strides : int, default 1
            Strides for each convolution layer inside the carndinal paths
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert dimension > 0
        assert cardinality > 0
        assert in_features % cardinality == 0 and out_features % cardinality == 0

        self.D = dimension
        num_input = in_features // cardinality
        num_output = out_features // cardinality

        self.dilations = []
        if dilations is None:
            # Default
            self.dilations = [3**i for i in range(cardinality)]
        elif isinstance(dilations, int):
            self.dilations = [dilations for _ in range(cardinality)]
        elif isinstance(dilations, list):
            assert len(dilations) == cardinality
            self.dilations = dilations
        else:
            raise ValueError("Invalid type for input strides, must be int or list!")

        self.kernels = []
        if isinstance(kernel_sizes, int):
            self.kernels = [kernel_sizes for _ in range(cardinality)]
        elif isinstance(kernel_sizes, list):
            assert len(kernel_sizes) == cardinality
            self.kernels = kernel_sizes
        else:
            raise ValueError("Invalid type for input strides, must be int or list!")

        self.strides = []
        if isinstance(strides, int):
            self.strides = [strides for _ in range(cardinality)]
        elif isinstance(strides, list):
            assert len(strides) == cardinality
            self.strides = strides
        else:
            raise ValueError("Invalid type for input strides, must be int or list!")

        # For each path, generate sequentials
        self.paths = []
        for i in range(cardinality):
            m = []
            m.append(ME.MinkowskiLinear(in_features, num_input))
            for j in range(depth):
                in_C = num_input if j == 0 else num_output
                m.append(
                    ME.MinkowskiConvolution(
                        in_channels=in_C,
                        out_channels=num_output,
                        kernel_size=self.kernels[i],
                        stride=self.strides[i],
                        dilation=self.dilations[i],
                        dimension=self.D,
                    )
                )
                m.append(norm_factory(normalization, num_output))
                m.append(act_factory(activation))
            m = nn.Sequential(*m)
            self.paths.append(m)
        self.paths = nn.Sequential(*self.paths)
        self.linear = ME.MinkowskiLinear(out_features, out_features)

        # Skip Connection
        if in_features != out_features:
            self.residual = ME.MinkowskiLinear(in_features, out_features)
        else:
            self.residual = nn.Identity()

    def forward(self, x):
        """Pass a tensor through the ResNeXt block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        residual = self.residual(x)

        cat = tuple([layer(x) for layer in self.paths])

        out = ME.cat(cat)
        out = self.linear(out)
        out += residual

        return out


class SPP(ME.MinkowskiNetwork):
    """Spatial Pyramid Pooling.

    PSPNet (Pyramid Scene Parsing Network) uses vanilla SPPs, while
    DeeplabV3 and DeeplabV3+ uses ASPP (atrous versions).

    Default parameters will construct a global average pooling + unpooling
    layer which is done in ParseNet.
    """

    def __init__(
        self,
        in_features,
        out_features,
        kernel_sizes=None,
        dilations=None,
        mode="avg",
        dimension=3,
    ):
        """Initialize the SPP block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        kernel_sizes : int, optional
            Kernel sizes for each pooling operation
        dilations : int, optional
            Dilation rates for atrous convolutions. Note that
            `kernel_size == stride` for an SPP layer.
        mode : str, default 'avg'
            Pooling mode (one of 'avg', 'max' and 'sum'
        dimension : int, default 3
            Dimension of the input image
        """
        # Initialize the parent class
        super().__init__(dimension)

        if mode == "avg":
            self.pool_fn = ME.MinkowskiAvgPooling
        elif mode == "max":
            self.pool_fn = ME.MinkowskiMaxPooling
        elif mode == "sum":
            self.pool_fn = ME.MinkowskiSumPooling
        else:
            raise ValueError("Invalid pooling mode, must be one of \
                'sum', 'max' or 'average'")

        self.unpool_fn = ME.MinkowskiPoolingTranspose

        # Include global pooling as first modules.
        self.pool = [ME.MinkowskiGlobalPooling(dimension=dimension)]
        self.unpool = [ME.MinkowskiBroadcast(dimension=dimension)]
        multiplier = 1

        # Define subregion poolings
        self.spp = []
        if kernel_sizes is not None:
            if isinstance(dilations, int):
                dilations = [dilations for _ in range(len(kernel_sizes))]
            elif isinstance(dilations, list):
                assert len(kernel_sizes) == len(dilations)
            else:
                raise ValueError(
                    "Invalid input to dilations, must be either " "int or list of ints."
                )
            multiplier = len(kernel_sizes) + 1  # Additional 1 for globalPool
            for k, d in zip(kernel_sizes, dilations):
                pooling_layer = self.pool_fn(
                    kernel_size=k, dilation=d, stride=k, dimension=dimension
                )
                unpooling_layer = self.unpool_fn(
                    kernel_size=k, dilation=d, stride=k, dimension=dimension
                )
                self.pool.append(pooling_layer)
                self.unpool.append(unpooling_layer)
        self.pool = nn.Sequential(*self.pool)
        self.unpool = nn.Sequential(*self.unpool)
        self.linear = ME.MinkowskiLinear(in_features * multiplier, out_features)

    def forward(self, input):
        """Pass a tensor through the SPP block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        cat = []
        for i, pool in enumerate(self.pool):
            x = pool(input)
            # First item is Global Pooling
            if i == 0:
                x = self.unpool[i](input, x)
            else:
                x = self.unpool[i](x)
            cat.append(x)
        out = ME.cat(cat)
        out = self.linear(out)

        return out


class ASPP(ME.MinkowskiNetwork):
    """Atrous Spatial Pyramid Pooling."""

    def __init__(
        self,
        in_features,
        out_features,
        dimension=3,
        width=5,
        dilations=[2, 4, 6, 8, 12],
    ):
        """Initialize the ASPP block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        dimension : int, default 3
            Dimension of the input image
        dilations : list, default [2, 4, 6, 8, 12]
            Dilation rates for atrous convolutions
        width : int, default 5
            Width of atrous convolutions
        """
        # Initialize parent class
        super().__init__(dimension)

        assert len(dilations) == width

        m = []
        m.append(ME.MinkowskiLinear(in_features, out_features))
        for d in dilations:
            m.append(
                ME.MinkowskiConvolution(
                    in_features,
                    out_features,
                    kernel_size=3,
                    dilation=d,
                    dimension=self.D,
                )
            )
        self.net = nn.Sequential(*m)
        self.pool = ME.MinkowskiGlobalPooling(dimension=self.D)
        self.unpool = ME.MinkowskiBroadcast(dimension=self.D)
        self.out = nn.Sequential(
            ME.MinkowskiConvolution(
                in_features * (2 + width),
                out_features,
                kernel_size=3,
                dilation=1,
                dimension=self.D,
            ),
            ME.MinkowskiBatchNorm(out_features),
            ME.MinkowskiReLU(),
        )

    def forward(self, x):
        """Pass a tensor through the ASPP block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        cat = []
        for i, layer in enumerate(self.net):
            x_i = layer(x)
            cat.append(x_i)
        x_global = self.pool(x)
        x_global = self.unpool(x, x_global)
        cat.append(x_global)
        out = ME.cat(cat)
        return self.out(out)


class CascadeDilationBlock(ME.MinkowskiNetwork):
    """Cascaded Atrous Convolution Block."""

    def __init__(
        self,
        in_features,
        out_features,
        dimension=3,
        depth=6,
        dilations=[1, 2, 4, 8, 16, 32],
        activation="relu",
    ):
        """Initialize the Cascaded Atrous Convolution block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        dimension : int, default 3
            Dimension of the input image
        depth : int, default 1
            Number of atrous convolutions layers
        dilations : list, default [1, 2, 4, 8, 16, 32]
            Dilation rates for atrous convolutions
        activation : union[str, dict], default 'relu'
            activation function configuration
        """
        # Initialize parent class
        super().__init__(dimension)

        F = out_features
        m = []
        self.input_layer = ME.MinkowskiLinear(in_features, F)
        for i in range(depth):
            m.append(ResNetBlock(F, F, dilation=dilations[i], activation=activation))
        self.net = nn.Sequential(*m)

    def forward(self, x):
        """Pass a tensor through the Cascaded Atrous Convolution block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        x = self.input_layer(x)
        sumTensor = x
        for i, layer in enumerate(self.net):
            x = layer(x)
            sumTensor += x

        return sumTensor


class MBConv(ME.MinkowskiNetwork):
    """MBConv block."""

    def __init__(
        self,
        in_features,
        out_features,
        expand_ratio=2,
        dimension=3,
        dilation=1,
        kernel_size=3,
        stride=1,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the MBConv block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        expand_ratio : int, default 2
            Multiplicative factor to apply to the input number of features
        dimension : int, default 3
            Dimension of the input image
        dilation : int, default 1
            Convolution kernel dilation
        kernel_size : int, default 3
            Convolution kernel size
        stride : int, default 1
            Convolution kernel stride
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(dimension)

        self.D = dimension
        self.hidden_dim = int(expand_ratio * in_features)

        if expand_ratio == 1:
            self.m = nn.Sequential(
                norm_factory(normalization, in_features),
                act_factory(activation),
                ME.MinkowskiConvolution(
                    in_features,
                    out_features,
                    kernel_size=3,
                    stride=1,
                    dilation=1,
                    dimension=self.D,
                    bias=bias,
                ),
            )
        else:
            self.m = nn.Sequential(
                norm_factory(normalization, in_features),
                act_factory(activation),
                ME.MinkowskiLinear(in_features, self.hidden_dim),
                norm_factory(normalization, self.hidden_dim),
                act_factory(activation),
                ME.MinkowskiChannelwiseConvolution(
                    self.hidden_dim,
                    kernel_size=kernel_size,
                    stride=stride,
                    dilation=dilation,
                    bias=bias,
                    dimension=self.D,
                ),
                norm_factory(normalization, self.hidden_dim),
                act_factory(activation),
                ME.MinkowskiLinear(self.hidden_dim, out_features),
            )

    def forward(self, x):
        """Pass a tensor through the MBConv block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        out = self.m(x)

        return out


class MBResConv(ME.MinkowskiNetwork):
    """MBResConv block."""

    def __init__(
        self,
        in_features,
        out_features,
        expand_ratio=2,
        dimension=3,
        dilation=1,
        kernel_size=3,
        stride=1,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the MBResConv block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        expand_ratio : int, default 2
            Multiplicative factor to apply to the input number of features
        dimension : int, default 3
            Dimension of the input image
        dilation : int, default 1
            Convolution kernel dilation
        kernel_size : int, default 3
            Convolution kernel size
        stride : int, default 1
            Convolution kernel stride
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(dimension)

        self.D = dimension
        self.m1 = MBConv(
            in_features,
            out_features,
            expand_ratio=expand_ratio,
            dimension=dimension,
            dilation=dilation,
            kernel_size=kernel_size,
            stride=stride,
            activation=activation,
            normalization=normalization,
            bias=bias,
        )
        self.m2 = MBConv(
            out_features,
            out_features,
            expand_ratio=expand_ratio,
            dimension=dimension,
            dilation=dilation,
            kernel_size=kernel_size,
            stride=stride,
            activation=activation,
            normalization=normalization,
            bias=bias,
        )
        if in_features == out_features:
            self.connection = nn.Identity()
        else:
            self.connection = nn.Sequential(
                norm_factory(normalization, in_features),
                act_factory(activation),
                ME.MinkowskiLinear(in_features, out_features),
            )

    def forward(self, x):
        """Pass a tensor through the MBResConv block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        x_add = self.connection(x)
        x = self.m1(x)
        x = self.m2(x)
        out = x_add + x

        return out


class SEBlock(ME.MinkowskiNetwork):
    """Squeeze and Excitation block."""

    def __init__(self, channels, ratio=8, dimension=3):
        """Initialize the SE block.

        Parameters
        ----------
        channel : int
            Number of input features
        ratio : int, default 8
            Squeezing ratio
        dimension : int, default 3
            Dimension of the input image
        """
        # Initialize the parent class
        super().__init__(dimension)

        assert channels // ratio > 0
        assert channels % ratio == 0
        self.linear1 = ME.MinkowskiLinear(channels, channels // ratio)
        self.relu = ME.MinkowskiReLU()
        self.linear2 = ME.MinkowskiLinear(channels // ratio, channels)
        self.sigmoid = ME.MinkowskiSigmoid()
        self.pool = ME.MinkowskiGlobalPooling()
        self.bcst = ME.MinkowskiBroadcastMultiplication()

    def forward(self, x):
        """Pass a tensor through the SE block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        g = self.pool(x)
        g = self.linear1(g)
        g = self.relu(g)
        g = self.linear2(g)
        g = self.sigmoid(g)
        out = self.bcst(x, g)

        return x


class SEResNetBlock(ME.MinkowskiNetwork):
    """Squeeze and Excitation ResNet block."""

    def __init__(
        self,
        in_features,
        out_features,
        se_ratio=8,
        stride=1,
        dilation=1,
        dimension=3,
        activation="relu",
        normalization="batch_norm",
    ):
        """Initialize the SEResNet block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        se_ratio : int, default 8
            Squeezing ratio
        stride : int, default 1
            Convolution kernel stride
        dilation : int, default 1
            Convolution kernel dilation
        dimension : int, default 3
            Dimension of the input image
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        """
        # Initialize parent class
        super().__init__(dimension)

        assert dimension > 0

        if in_features != out_features:
            self.residual = ME.MinkowskiLinear(in_features, out_features)
        else:
            self.residual = nn.Identity()

        self.conv1 = ME.MinkowskiConvolution(
            in_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
        )
        self.norm1 = norm_factory(normalization, out_features)
        self.act_fn1 = act_factory(activation)

        self.conv2 = ME.MinkowskiConvolution(
            out_features,
            out_features,
            kernel_size=3,
            stride=1,
            dilation=dilation,
            dimension=dimension,
        )
        self.norm2 = norm_factory(normalization, out_features)
        self.act_fn2 = act_factory(activation)

        self.se_block = SEBlock(out_features, ratio=se_ratio, dimension=dimension)

    def forward(self, x):
        """Pass a tensor through the SEResNet block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        residual = self.residual(x)
        out = self.act_fn1(self.norm1(self.conv1(x)))
        out = self.norm2(self.conv2(out))
        out = self.se_block(out)
        out += residual
        out = self.act_fn2(out)

        return out


class MBResConvSE(MBResConv):
    """MBResConvSE block."""

    def __init__(
        self,
        in_features,
        out_features,
        se_ratio=8,
        expand_ratio=2,
        dimension=3,
        dilation=1,
        kernel_size=3,
        stride=1,
        activation="relu",
        normalization="batch_norm",
        bias=False,
    ):
        """Initialize the MBResConvSE block.

        Parameters
        ----------
        in_features : int
            Number of input features
        out_features : int
            Number of output features
        se_ratio : int, default 8
            Squeezing ratio
        expand_ratio : int, default 2
            Multiplicative factor to apply to the input number of features
        dimension : int, default 3
            Dimension of the input image
        dilation : int, default 1
            Convolution kernel dilation
        kernel_size : int, default 3
            Convolution kernel size
        stride : int, default 1
            Convolution kernel stride
        activation : union[str, dict], default 'relu'
            activation function configuration
        normalization : union[str, dict], default 'batch_norm'
            normalization function configuration
        bias : bool, default False
            Whether to add a bias term to the kernel
        """
        # Initialize the parent class
        super().__init__(
            in_features,
            out_features,
            expand_ratio=expand_ratio,
            dimension=dimension,
            dilation=dilation,
            kernel_size=kernel_size,
            stride=stride,
            activation=activation,
            normalization=normalization,
            bias=bias,
        )

        if in_features == out_features:
            self.connection = nn.Identity()
        else:
            self.connection = nn.Sequential(
                norm_factory(normalization, in_features),
                act_factory(activation),
                ME.MinkowskiLinear(in_features, out_features),
            )

        self.se = SEBlock(out_features, ratio=se_ratio)

    def forward(self, x):
        """Pass a tensor through the SEResNet block.

        Parameters
        ----------
        x : ME.SparseTensor
            Input sparse tensor

        Returns
        -------
        ME.SparseTensor
            Output sparse tensor
        """
        res = super().forward(x)
        attn = self.se(res)
        out = self.connection(x) + attn

        return out



"""Module which defines generic graph features update based on MLPs."""

import torch
from torch import nn
from torch_geometric.utils import softmax
from torch_scatter import scatter

from spine.model.layer.common.mlp import MLP

__all__ = ["MLPEdgeLayer", "MLPNodeLayer", "MLPGlobalLayer"]


class MLPEdgeLayer(nn.Module):
    """Model used to update the set of edge features.

    For each edge, this model first aggregates the features from the source
    node (N_c) with those of the sink node (N_c) and those of the edge (N_e)
    and those of the graph (N_g) to form an input feature vector of dimension
    (E, 2*N_c + N_e + N_g). This feature vector is then passed through a
    multi-layer perceptron (MLP) and outputs an (E, N_o) vector, with N_o the
    width of the MLP (feature size of the hidden representation).
    """

    # Name of the edge layer (as specified in the configuration)
    name = "mlp"

    def __init__(self, node_in, edge_in, glob_in, mlp):
        """Initialize the MLP which is used to update the edge features.

        Parameters
        ----------
        edge_in : int
            Number of input edge features
        node_in : int
            Number of input node features
        glob_in : int
            Number of input global features for the graph
        mlp : dict
            Configuration of the edge update MLP
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the edge update MLP
        self.mlp = MLP(2 * node_in + edge_in + glob_in, **mlp)
        self.feature_size = self.mlp.feature_size

    def forward(self, src_feats, dest_feats, edge_feats, glob_feats, batch):
        """Pass a batch of node/edges through the edge update layer.

        Parameters
        ----------
        src_feats : torch.Tensor
            (E, N_c) Source node features
        dest_feats : torch.Tensor
            (E, N_c) Sink node features
        edge_feats : torch.Tensor
            (E, N_e) Edge features
        glob_feats : torch.Tensor
            (B, N_g) Global features
        batch : torch.Tensor
            (E) ID of the entry of each of the edges within the batch

        Returns
        -------
        torch.Tensor
            (C, N_o) Updated edge features
        """
        # Initialize the input to the edge update
        input_feats = torch.cat([src_feats, dest_feats, edge_feats], dim=1)
        if glob_feats is not None:
            input_feats = torch.cat([input_feats, glob_feats[batch]], dim=1)

        # Pass those features through the MLP
        return self.mlp(input_feats)


class MLPNodeLayer(nn.Module):
    """Model used to update the set of node features.

    For each node, this model proceeds in two seperate steps:
    - A message formation step
    - A message aggregation step

    For each edge, the message formation step consists in first aggregating
    the source node features (N_c) with the edge features (N_e) to form an
    input feature vector of dimension (N_c + N_e). This feature vector is
    then passed through a multi-layer perceptron (MLP) and outputs an (N_o)
    vector, with N_o the width of the MLP (feature size of the hidden
    representation). This feature vector is the message associated with that
    edge.

    For each node, the message aggregations step consists in taking a
    summary statistic of all edge features which correspond to edges that have
    that node as a sink. The so-formed feature vector of size (N_o) is then
    stacked with the node features (N_c) and the global graph features (N_g)
    to form a (N_o + N_c + N_g) feature vector. This new vector is passed
    through a second MLP to update the node features to (N_o').
    """

    # Name of the node layer (as specified in the configuration)
    name = "mlp"

    def __init__(
        self,
        node_in,
        edge_in,
        glob_in,
        message_mlp,
        aggr_mlp,
        reduction="mean",
        attention=False,
    ):
        """Initialize the MLPs which are used to update the node features.

        Parameters
        ----------
        node_in : int
            Number of input node features
        edge_in : int
            Number of input edge features
        glob_in : int
            Number of input global features for the graph
        message_mlp : dict
            Configuration of the message creation MLP
        aggr_mlp : dict
            Configuration of the message aggregtion MLP
        reduction : str, default 'mean'
            Message feature aggregation function
        attention : bool, default False
            Whether or not to learn explicit attention to particular messages
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the message creation MLP
        self.message_mlp = MLP(node_in + edge_in, **message_mlp)
        message_width = self.message_mlp.width[-1]

        # Initialize the aggregation MLP
        self.aggr_mlp = MLP(node_in + message_width + glob_in, **aggr_mlp)
        self.feature_size = self.aggr_mlp.feature_size

        # Store the edge feature aggregation mode
        self.reduction = reduction

        # If self-attention is needed, initialize a simple gate layer
        self.attention = attention
        if self.attention:
            self.gate = nn.Linear(node_out, 1)

    def forward(self, node_feats, edge_index, edge_feats, glob_feats, batch):
        """Pass a batch of node/edges through the edge update layer.

        Parameters
        ----------
        node_feats : torch.Tensor
            (C, N_c) Node features
        edge_index : torch.Tensor
            (2, E) Incidence matrix
        edge_feats : torch.Tensor
            (E, N_e) Edge features
        glob_feats : torch.Tensor
            (B, N_g) Global features
        batch : torch.Tensor
            (C) ID of the entry of each of the nodes within the batch

        Returns
        -------
        torch.Tensor
            (C, N_o') Updated node features
        """
        # Slip the edge index between source and sink
        src, dest = edge_index

        # Concatenate edge features with the source nodes, pass through MLP
        input_feats = torch.cat([node_feats[src], edge_feats], dim=1)
        message_feats = self.message_mlp(input_feats)

        # If requested, make the network learn to explicitely tune down
        # features associated with certaine edges
        if self.attention:
            # Compute a softmax score for nodes that share a sink
            weights = softmax(self.gate(message_feats), index=dest)

            # Apply the gate weigths to the message features
            message_feats = weights * message_feats

        # Aggregate messages to form one coherent message per node
        message_feats = scatter(
            message_feats,
            dest,
            dim=0,
            dim_size=node_feats.size(0),
            reduce=self.reduction,
        )

        # Aggerate the message with the node features, pass through MLP
        input_feats = torch.cat([node_feats, message_feats], dim=1)
        if glob_feats is not None:
            input_feats = torch.cat([input_feats, glob_feats[batch]], dim=1)

        return self.aggr_mlp(input_feats)


class MLPGlobalLayer(nn.Module):
    """Model used to update the set of global graph features.

    For each graph (one per entry in the batch of B graphs), this model first
    takes a summary statistic of the information from all the nodes in the
    graph (N_c). It then aggregates this feature vector with the one associated
    with the graph itself (N_g) to form a feature fector of dimension
    (N_c + N_g). This feature vector is then passed through a multi-layer
    perceptron (MLP) and outputs a (B, N_o) vector, with N_o the width of the
    MLP (feature size of the hidden representation).
    """

    # Name of the global layer (as specified in the configuration)
    name = "mlp"

    def __init__(self, node_in, glob_in, mlp, reduction="mean"):
        """Initialize the MLP which is used to update the edge features.

        Parameters
        ----------
        glob_in : int
            Number of input global features for the graph
        node_in : int
            Number of input node features
        mlp : dict
            Configuration of the global representation update MLP
        reduction : str, default 'mean'
            Message feature aggregation function
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the global representation update MLP
        self.mlp = MLP(node_in + glob_in, **mlp)
        self.feature_size = self.mlp.feature_size

        # Store the node feature aggregation mode
        self.reduction = reduction

    def forward(self, node_feats, edge_index, edge_feats, glob_feats, batch):
        """Pass a batch of node/edges through the global update layer.

        Parameters
        ----------
        node_feats : torch.Tensor
            (C, N_c) Source node features
        edge_index : torch.Tensor
            (2, E) Incidence matrix
        edge_feats : torch.Tensor
            (E, N_e) Edge features
        glob_feats : torch.Tensor
            (B, N_g) Global features
        batch : torch.Tensor
            (C) ID of the entry of each of the nodes within the batch

        Returns
        -------
        torch.Tensor
            (B, N_o) Updated global features
        """
        # Aggregate node features to form one single feature vector
        all_node_feats = scatter(node_feats, batch, dim=0, reduce=self.reduction)

        # Initialize the input to the global update
        input_feats = torch.cat([all_node_feats, glob_feats], dim=1)

        # Pass those features through the MLP
        return self.mlp(input_feats)



import MinkowskiEngine as ME
import MinkowskiFunctional as MF
import numpy as np
import torch
import torch.nn as nn

from spine.model.layer.cnn.blocks import ResNetBlock
from spine.model.layer.cnn.configuration import setup_cnn_configuration
from spine.model.layer.factories import loss_fn_factory
from spine.utils.gnn.cluster import form_clusters, get_cluster_label

from .act_norm import act_factory
from .ppn import *


class VertexPPN(nn.Module):

    def __init__(self, cfg, name="vertex_ppn"):
        super(VertexPPN, self).__init__()
        setup_cnn_configuration(self, cfg, name)

        self.model_cfg = cfg.get(name, {})
        # UResNet Configurations
        self.reps = self.model_cfg.get("reps", 2)
        self.depth = self.model_cfg.get("depth", 5)
        self.num_filters = self.model_cfg.get("filters", 16)
        self.nPlanes = [i * self.num_filters for i in range(1, self.depth + 1)]
        self.vertex_score_threshold = self.model_cfg.get("score_threshold", 0.5)
        self.input_kernel = self.model_cfg.get("input_kernel", 3)

        # Initialize Decoder
        self.decoding_block = []
        self.decoding_conv = []
        self.vertex_pred = nn.ModuleList()
        for i in range(self.depth - 2, -1, -1):
            m = []
            m.append(ME.MinkowskiBatchNorm(self.nPlanes[i + 1]))
            m.append(act_factory(self.activation_name, **self.activation_args))
            m.append(
                ME.MinkowskiConvolutionTranspose(
                    in_channels=self.nPlanes[i + 1],
                    out_channels=self.nPlanes[i],
                    kernel_size=2,
                    stride=2,
                    dimension=self.D,
                )
            )
            m = nn.Sequential(*m)
            self.decoding_conv.append(m)
            m = []
            for j in range(self.reps):
                m.append(
                    ResNetBlock(
                        self.nPlanes[i] * (2 if j == 0 else 1),
                        self.nPlanes[i],
                        dimension=self.D,
                        activation=self.activation_name,
                        activation_args=self.activation_args,
                    )
                )
            m = nn.Sequential(*m)
            self.decoding_block.append(m)
            self.vertex_pred.append(ME.MinkowskiLinear(self.nPlanes[i], 1))
        self.decoding_block = nn.Sequential(*self.decoding_block)
        self.decoding_conv = nn.Sequential(*self.decoding_conv)

        self.sigmoid = ME.MinkowskiSigmoid()
        self.expand_as = ExpandAs()

        self.final_block = ResNetBlock(
            self.nPlanes[0],
            self.nPlanes[0],
            dimension=self.D,
            activation=self.activation_name,
            activation_args=self.activation_args,
        )

        self.vertex_regression = ME.MinkowskiConvolution(
            self.nPlanes[0], self.D, kernel_size=3, stride=1, dimension=self.D
        )
        self.vertexness_score = ME.MinkowskiConvolution(
            self.nPlanes[0], 2, kernel_size=3, stride=1, dimension=self.D
        )

    def forward(
        self,
        final,
        decoderTensors,
        input_sparse_tensor=None,
        primary_labels=None,
        segment_labels=None,
    ):
        vertex_layers, vertex_coords = [], []
        tmp = []
        segment_label_scales = []
        primary_label_scales = []
        device = final.device

        # We need to make labels on-the-fly to include true points in the
        # propagated masks during training

        decoder_feature_maps = decoderTensors

        x = final

        for i, layer in enumerate(self.decoding_conv):

            decTensor = decoder_feature_maps[i]
            x = layer(x)
            x = ME.cat(decTensor, x)
            x = self.decoding_block[i](x)
            scores = self.vertex_pred[i](x)
            tmp.append(scores.F)

            vertex_coords.append(scores.C)
            scores = self.sigmoid(scores)

            s_expanded = self.expand_as(scores, x.F.shape)
            x = x * s_expanded.detach()

        device = x.F.device
        vertex_output_coordinates = x.C
        for p in tmp:
            a = p.to(dtype=torch.float32, device=device)
            vertex_layers.append(a)

        x = self.final_block(x)
        pixel_pred = self.vertex_regression(x)
        vertex_final_score = self.vertexness_score(x)

        # X, Y, Z, logits
        points = torch.cat([pixel_pred.F, vertex_final_score.F], dim=1)

        if primary_labels is not None:

            primary_labels_curr = primary_labels
            segment_labels_curr = segment_labels

            for i, x in enumerate(reversed(decoderTensors[:3])):
                print(i, x.tensor_stride, decoderTensors[::-1][i].tensor_stride)

                primary_label_layer = torch.zeros(
                    x.C.shape[0], dtype=torch.bool, device=x.F.device
                )
                segment_label_layer = torch.zeros(
                    x.C.shape[0], dtype=torch.long, device=x.F.device
                )

                kernel_map = input_sparse_tensor.coordinate_manager.kernel_map(
                    decoderTensors[::-1][i].tensor_stride, x.tensor_stride
                )

                for k, curr in kernel_map.items():
                    primary_label_layer[curr[1].long()] = primary_labels_curr[
                        curr[0].long()
                    ].bool()
                    segment_label_layer[curr[1].long()] = segment_labels_curr[
                        curr[0].long()
                    ].long()

                primary_label_scales.append(primary_label_layer)
                segment_label_scales.append(segment_label_layer)

                primary_labels_curr = primary_label_layer
                segment_labels_curr = segment_label_layer

        res = {
            "vertex_points": [points],
            "vertex_layers": [vertex_layers],
            "vertex_coords": [vertex_coords],
            "vertex_output_coordinates": [vertex_output_coordinates],
        }

        if primary_labels is not None:
            res["primary_label_scales"] = [primary_label_scales[::-1]]
            res["segment_label_scales"] = [segment_label_scales[::-1]]

        return res


class VertexPPNLoss(torch.nn.modules.loss._Loss):
    """
    Loss function for PPN.

    Output
    ------
    vertex_reg_loss : float
        Distance loss
    vertex_mask_loss : float
        Binary voxel-wise prediction (is there an object of interest or not)
    vertex_loss : float
        Combined loss
    vertex_accuracy : float
        Combined accuracy

    See Also
    --------
    PPN, spine.model.uresnet_ppn_chain
    """

    def __init__(self, cfg, name="ppn"):
        super(VertexPPNLoss, self).__init__()
        self.loss_config = cfg.get(name, {})
        # pprint(self.loss_config)
        self.mask_loss = self.loss_config.get("mask_loss", "bce")
        self.lossfn = loss_fn_factory(self.mask_loss)
        self.resolution = self.loss_config.get("ppn_resolution", 1.0)
        self.regloss = torch.nn.MSELoss()
        self.use_numpy = self.loss_config.get("use_numpy", False)

    def get_vertices(self, kinematics_label):

        data = kinematics_label[0].cpu().numpy()
        batch_ids = data[:, 0]
        clusts = form_clusters(data, column=7)
        inter_label = get_cluster_label(data, clusts, column=7)
        batch_index = get_cluster_batch(data, clusts)
        primary_labels = get_cluster_label(data, clusts, column=8)
        vtx_x = get_cluster_label(data, clusts, column=10)
        vtx_y = get_cluster_label(data, clusts, column=11)
        vtx_z = get_cluster_label(data, clusts, column=12)

        vertices_label = np.hstack(
            [
                batch_index.reshape(-1, 1),
                vtx_x.reshape(-1, 1),
                vtx_y.reshape(-1, 1),
                vtx_z.reshape(-1, 1),
                inter_label.reshape(-1, 1),
                primary_labels.reshape(-1, 1),
            ]
        )

        return [vertices_label]

    def compute_heatmap(self, coords_batch, vertices_batch, primaries_batch):

        heatmap = torch.zeros(
            coords_batch.shape[0], dtype=torch.float32, device=coords_batch.device
        )

        primary_mask = primaries_batch == 1
        heatmap[primary_mask] = 1.0

        valid_coords = coords_batch[primary_mask]

        print(heatmap, valid_coords.shape, vertices_batch.shape)
        print(vertices_batch)
        assert False
        return 0

    def forward(self, result, kinematics_label):

        batch_ids = [result["vertex_coords"][0][-1][:, 0]]
        num_batches = len(batch_ids[0].unique())
        total_loss = 0
        total_acc = 0
        device = kinematics_label[0].device

        res = {"vertex_reg_loss": 0.0, "vertex_mask_loss": 0.0}

        particles_label = self.get_vertices(kinematics_label)

        print("Particles Label = ", particles_label[0], type(particles_label))

        # Semantic Segmentation Loss
        for igpu in range(len(kinematics_label)):
            particles = particles_label[igpu]
            ppn_layers = result["vertex_layers"][igpu]
            ppn_coords = result["vertex_coords"][igpu]
            points = result["vertex_points"][igpu]
            primary_label_scales = result["primary_label_scales"][igpu]
            loss_gpu, acc_gpu = 0.0, 0.0
            num_layers = len(ppn_layers)
            for layer in range(len(ppn_layers)):
                # print("Layer = ", layer)
                ppn_score_layer = ppn_layers[layer]
                coords_layer = ppn_coords[layer]
                primaries_layer = primary_label_scales[layer]
                loss_layer = 0.0
                for b in batch_ids[igpu].int().unique():

                    batch_index_layer = coords_layer[:, 0].int() == b
                    batch_particle_index = batch_ids[igpu].int() == b
                    print(particles, particles[:, 0], b)
                    print("particles[:, 0] == b = ", particles[:, 0] == b.item())
                    points_label_batch = particles[particles[:, 0] == b]
                    vertices_batch = points_label_batch[:, 1:4]
                    print("points_label_batch = ", points_label_batch)
                    scores_event = ppn_score_layer[batch_index_layer].squeeze()
                    coords_batch = coords_layer[batch_index_layer]
                    primaries_batch = primaries_layer[batch_index_layer]

                    heatmap = self.compute_heatmap(
                        coords_batch, vertices_batch, primaries_batch
                    )

            #         d_true = pairwise_distances(
            #             points_label,
            #             points_event[:, 1:4].to(device))

            #         d_positives = (d_true < self.resolution * \
            #                        2**(len(ppn_layers) - layer)).any(dim=0)

            #         num_positives = d_positives.sum()
            #         num_negatives = d_positives.nelement() - num_positives

            #         w = num_positives / \
            #             (num_positives + num_negatives)

            #         weight_ppn = torch.zeros(d_positives.shape[0]).to(device)
            #         weight_ppn[d_positives] = 1 - w
            #         weight_ppn[~d_positives] = w

            #         loss_batch = self.lossfn(scores_event,
            #                                  d_positives,
            #                                  weight=weight_ppn,
            #                                  reduction='mean')

            #         loss_layer += loss_batch
            #         if layer == len(ppn_layers)-1:

            #             # Get Final Layers
            #             anchors = coords_layer[batch_particle_index][:, 1:4].to(device) + 0.5
            #             pixel_score = points[batch_particle_index][:, -1]
            #             pixel_logits = points[batch_particle_index][:, 3:8]
            #             pixel_pred = points[batch_particle_index][:, :3] + anchors

            #             d = pairwise_distances(points_label, pixel_pred)
            #             positives = (d < self.resolution).any(dim=0)
            #             if (torch.sum(positives) < 1):
            #                 continue
            #             acc = (positives == (pixel_score > 0)).sum() / float(pixel_score.shape[0])
            #             total_acc += acc

            #             # Mask Loss
            #             mask_loss_final = self.lossfn(pixel_score,
            #                                           positives,
            #                                           weight=weight_ppn,
            #                                           reduction='mean')

            #             distance_positives = d[:, positives]

            #             # Distance Loss
            #             d2, _ = torch.min(distance_positives, dim=0)
            #             reg_loss = d2.mean()
            #             res['vertex_reg_loss'] += float(reg_loss) / num_batches
            #             res['vertex_mask_loss'] += float(mask_loss_final) / num_batches
            #             total_loss += (reg_loss + mask_loss_final) / num_batches

            #     loss_layer /= num_batches
            #     loss_gpu += loss_layer
            # loss_gpu /= len(ppn_layers)
            # total_loss += loss_gpu

        total_acc /= num_batches
        res["vertex_loss"] = total_loss
        res["vertex_accuracy"] = float(total_acc)
        return res



"""Factories to build the GNN model components."""

from spine.utils.factory import instantiate, module_dict

from . import encode, graph, loss, model

__all__ = [
    "graph_factory",
    "gnn_model_factory",
    "node_encoder_factory",
    "edge_encoder_factory",
    "global_encoder_factory",
    "node_loss_factory",
    "edge_loss_factory",
    "global_loss_factory",
]


def graph_factory(cfg, classes):
    """Instantiates a graph constructor from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Graph constructor configuration
    classes : Union[int, list]
        List of classes to build a graph on

    Returns
    -------
    object
        Instantiated graph constructor
    """
    graph_dict = module_dict(graph)
    return instantiate(graph_dict, cfg, classes=classes)


def gnn_model_factory(cfg, node_pred, edge_pred, global_pred):
    """Instantiates a GNN model from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        GNN model configuration
    node_pred : bool
        Whether the model should return node features or not
    edge_pred : bool
        Whether the model should return edge features or not
    global_pred : bool
        Whether the model should return global features or not

    Returns
    -------
    object
        Instantiated GNN model
    """
    gnn_model_dict = module_dict(model)
    return instantiate(
        gnn_model_dict,
        cfg,
        node_pred=node_pred,
        edge_pred=edge_pred,
        global_pred=global_pred,
    )


def node_encoder_factory(cfg):
    """Instantiates a node encoder from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Node encoder configuration

    Returns
    -------
    object
        Instantiated node encoder
    """
    node_encoder_dict = module_dict(encode, pattern="Node")
    return instantiate(node_encoder_dict, cfg)


def edge_encoder_factory(cfg):
    """Instantiates an edge encoder from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Edge encoder configuration

    Returns
    -------
    object
        Instantiated edge encoder
    """
    edge_encoder_dict = module_dict(encode, pattern="Edge")
    return instantiate(edge_encoder_dict, cfg)


def global_encoder_factory(cfg):
    """Instantiates a global graph encoder from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Global graph encoder configuration

    Returns
    -------
    object
        Instantiated global graph encoder
    """
    global_encoder_dict = module_dict(encode, pattern="Global")
    return instantiate(global_encoder_dict, cfg)


def node_loss_factory(cfg):
    """Instantiates a node loss from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Node loss configuration

    Returns
    -------
    object
        Instantiated node loss
    """
    node_loss_dict = module_dict(loss, pattern="Node")
    return instantiate(node_loss_dict, cfg)


def edge_loss_factory(cfg):
    """Instantiates an edge loss from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Edge loss configuration

    Returns
    -------
    object
        Instantiated edge loss
    """
    edge_loss_dict = module_dict(loss, pattern="Edge")
    return instantiate(edge_loss_dict, cfg)


def global_loss_factory(cfg):
    """Instantiates a global graph loss from a configuration dictionary.

    Parameters
    ----------
    cfg : dict
        Global graph loss configuration

    Returns
    -------
    object
        Instantiated global graph loss
    """
    global_loss_dict = module_dict(loss, pattern="Global")
    return instantiate(global_loss_dict, cfg)



"""PyTorch utility scripts and extensions for enhanced functionality.

This module provides local extensions and utility functions that enhance
or work around limitations in PyTorch's built-in functionality.
"""

from ..conditional import TORCH_AVAILABLE, torch

__all__ = ["cdist_fast"]


def cdist_fast(v1, v2, metric="euclidean"):
    """Computes the pairwise distances between two `torch.Tensor` objects.

    This is necessary because the torch.cdist implementation is either
    slower (with the `donot_use_mm_for_euclid_dist` option) or produces
    dramatically wrong answers under certain situations (with the
    `use_mm_for_euclid_dist_if_necessary` option).

    Parameters
    ----------
    v1 : torch.Tensor
        (N, D) tensor of coordinates
    v2 : torch.Tensor
        (M, D) tensor of coordinates
    metric : str
        Distance metric

    Returns
    -------
    torch.Tensor
        (N, M) tensor of pairwise distances
    """
    if not TORCH_AVAILABLE:
        raise ImportError(
            "PyTorch is required for cdist_fast functionality. "
            "Install with: pip install spine-ml[model]"
        )

    v1_2 = v1.unsqueeze(1).expand(v1.size(0), v2.size(0), v1.size(1))
    v2_2 = v2.unsqueeze(0).expand(v1.size(0), v2.size(0), v1.size(1))
    if metric == "cityblock":
        return torch.abs(v2_2 - v1_2).sum(2)
    elif metric == "euclidean":
        return torch.sqrt(torch.pow(v2_2 - v1_2, 2).sum(2))
    elif metric == "chebyshev":
        return torch.abs(v2_2 - v1_2).amax(2)
